from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap
from arcgis.apps.hub.sites import SiteManager, Site
from collections import OrderedDict
from datetime import datetime
import json


def _lazy_property(fn):
    """Decorator that makes a property lazy-evaluated."""
    # http://stevenloria.com/lazy-evaluated-properties-in-python/
    attr_name = "_lazy_" + fn.__name__

    @property
    def _lazy_property(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _lazy_property


class Initiative(OrderedDict):
    """
    Represents an initiative within a Hub. An Initiative supports
    policy- or activity-oriented goals through workflows, tools and team collaboration.
    """

    def __init__(self, gis, initiativeItem):
        """
        Constructs an empty Initiative object
        """
        self.item = initiativeItem
        self._gis = gis
        self._hub = gis.hub
        try:
            self._initiativedict = self.item.get_data()
            pmap = PropertyMap(self._initiativedict)
            self.definition = pmap
        except:
            self.definition = None

    def __repr__(self):
        return '<%s title:"%s" owner:%s>' % (
            type(self).__name__,
            self.title,
            self.owner,
        )

    @property
    def itemid(self) -> str:
        """
        Returns the item id of the initiative item
        """
        return self.item.id

    @property
    def title(self) -> str:
        """
        Returns the title of the initiative item
        """
        return self.item.title

    @property
    def description(self) -> str:
        """
        Returns the initiative description
        """
        return self.item.description

    @property
    def snippet(self) -> str:
        """
        Getter/Setter for the initiative snippet
        """
        return self.item.snippet

    @snippet.setter
    def snippet(self, value):
        self.item.snippet = value

    @property
    def owner(self) -> str:
        """
        Returns the owner of the initiative item
        """
        return self.item.owner

    @property
    def tags(self) -> str:
        """
        Returns the tags of the initiative item
        """
        return self.item.tags

    @property
    def url(self) -> str:
        """
        Returns the url of the initiative site
        """
        try:
            return self.item.properties["url"]
        except:
            return self.item.url

    @property
    def site_id(self) -> str:
        """
        Returns the item id of the initiative site
        """
        try:
            return self.item.properties["siteId"]
        except:
            return self._initiativedict["steps"][0]["itemIds"][0]

    @property
    def site_url(self) -> str:
        """
        Getter/Setter for the url of the initiative site
        """
        try:
            return self.item.url
        except:
            return self.sites.get(self.site_id).url

    @site_url.setter
    def site_url(self, value):
        self.item.url = value

    @property
    def content_group_id(self) -> str:
        """
        Returns the group id for the content group
        """
        return self.item.properties["contentGroupId"]

    @property
    def collab_group_id(self) -> str:
        """
        Getter/Setter for the group id for the collaboration group
        """
        try:
            return self.item.properties["collaborationGroupId"]
        except:
            return None

    @collab_group_id.setter
    def collab_group_id(self, value):
        self.item.properties["collaborationGroupId"] = value

    @property
    def followers_group_id(self) -> str:
        """
        Returns the group id for the followers group
        """
        return self.item.properties["followersGroupId"]

    @_lazy_property
    def sites(self) -> SiteManager:
        """
        The resource manager for an Initiative's sites.
        See :class:`~hub.sites.SiteManager`.
        """
        return SiteManager(self._hub, self)

    @_lazy_property
    def all_events(self):
        """
        Fetches all events (past or future) pertaining to an initiative
        """
        return self._gis.hub.events.search(initiative_id=self.item.id)

    @_lazy_property
    def followers(self) -> list:
        """
        Fetches the list of followers for initiative.
        """
        # Fetch followers group
        _followers_group = self._gis.groups.get(self.followers_group_id)
        return _followers_group.get_members()

    def add_content(self, items_list: list):
        """
        Adds a batch of items to the initiative content library.

        =====================    ====================================================================
        **Parameter**             **Description**
        ---------------------    --------------------------------------------------------------------
        items_list               Required list. A list of Item or item ids to add to the initiative.
        =====================    ====================================================================

        """
        # If input list is of item_ids, generate a list of corresponding items
        if type(items_list[0]) == str:
            items = [self._gis.content.get(item_id) for item_id in items_list]
        else:
            items = items_list
        # Fetch existing sharing privileges for each item, to retain them after adding to content library
        for item in items:
            sharing = item.shared_with
            everyone = sharing["everyone"]
            org = sharing["org"]
            groups = sharing["groups"]
            # add current initiative's content group to list of groups to share to
            groups.append(self.content_group_id)
            # share item to this group
            status = item.share(everyone=everyone, org=org, groups=groups)
            if status["results"][0]["success"] == False:
                return status
        return status

    def delete(self) -> bool:
        """
        Deletes the initiative, its site and associated groups.
        If unable to delete, raises a RuntimeException.

        :return:
            A bool containing True (for success) or False (for failure).

        .. code-block:: python

            USAGE EXAMPLE: Delete an initiative successfully

            initiative1 = myHub.initiatives.get('itemId12345')
            initiative1.delete()

            >> True
        """
        if self.item is not None:
            # Fetch initiative site
            _site = self._gis.hub.sites.get(self.site_id)
            Site.delete(_site)
            # Fetch and delete Initiative Collaboration group if exists
            try:
                _collab_group = self._gis.groups.get(self.collab_group_id)
                _collab_group.protected = False
                _collab_group.delete()
            except:
                pass
            # Fetch Content Group and delete
            try:
                _content_group = self._gis.groups.get(self.content_group_id)
                _content_group.protected = False
                _content_group.delete()
            except:
                pass
            # Fetch Followers Group and delete
            try:
                _followers_group = self._gis.groups.get(self.followers_group_id)
                _followers_group.protected = False
                _followers_group.delete()
            except:
                pass
            # Delete initiative
            return self.item.delete()

    def reassign_to(self, target_owner: str):
        """
        Allows the administrator to reassign the initiative object from one
        user to another.

        .. note::
            This will transfer ownership of all items (site, pages, content) and groups that
            belong to this initiative to the new target_owner.

        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        target_owner              Required string. The new desired owner of the initiative.
        =====================     ====================================================================
        """
        # check if admin user is performing this action
        if "admin" not in self._gis.users.me.role:
            return Exception(
                "You do not have the administrator privileges to perform this action."
            )
        # check if core team is needed by checking the role of the target_owner
        if self._gis.users.get(target_owner).role == "org_admin":
            # check if the initiative comes with core team by checking owner's role
            if self._gis.users.get(self.owner).role == "org_admin":
                # fetch the core team for the initative
                core_team = self._gis.groups.get(self.collab_group_id)
                # fetch the contents shared with this team
                core_team_content = core_team.content()
                # check if target_owner is part of core team, else add them to core team
                members = core_team.get_members()
                if (
                    target_owner not in members["admins"]
                    or target_owner not in members["users"]
                ):
                    core_team.add_users(target_owner)
                # remove items from core team
                self._gis.content.unshare_items(core_team_content, groups=[core_team])
                # reassign to target_owner
                for item in core_team_content:
                    item.reassign_to(target_owner)
                # fetch the items again since they have been reassigned
                new_content_list = []
                for item in core_team_content:
                    item_temp = self._gis.content.get(item.id)
                    new_content_list.append(item_temp)
                # share item back to the content group
                self._gis.content.share_items(
                    new_content_list, groups=[core_team], allow_members_to_edit=True
                )
                # reassign core team to target owner
                core_team.reassign_to(target_owner)
            else:
                # create core team necessary for the initiative
                _collab_group_title = title + " Core Team"
                _collab_group_dict = {
                    "title": _collab_group_title,
                    "tags": [
                        "Hub Group",
                        "Hub Initiative Group",
                        "Hub Site Group",
                        "Hub Core Team Group",
                        "Hub Team Group",
                    ],
                    "access": "org",
                    "capabilities": "updateitemcontrol",
                    "membershipAccess": "collaboration",
                    "snippet": "Members of this group can create, edit, and manage the site, pages, and other content related to hub-groups.",
                }
                collab_group = self._gis.groups.create_from_dict(_collab_group_dict)
                collab_group.protected = True
                self.collab_group_id = collab_group.id
        else:
            # reassign the initiative, site, page items
            self.item.reassign_to(target_owner)
            site = self._hub.sites.get(self.site_id)
            site.item.reassign_to(target_owner)
            site_pages = site.pages.search()
            # If pages exist
            if len(site_pages) > 0:
                for page in site_pages:
                    # Unlink page (deletes if)
                    page.item.reassign_to(target_owner)
        # fetch content group
        content_team = self._gis.groups.get(self.content_group_id)
        # reassign to target_owner
        content_team.reassign_to(target_owner)
        # If it is a Hub Premium initiative, repeat for followers group
        if self._hub._hub_enabled:
            followers_team = self._gis.groups.get(self.followers_group_id)
            followers_team.reassign_to(target_owner)
        return self._gis.content.get(self.itemid)

    def share(
        self, everyone=False, org=False, groups=None, allow_members_to_edit=False
    ):
        """
        Shares an initiative and associated site with the specified list of groups.

        ======================  ========================================================
        **Parameter**            **Description**
        ----------------------  --------------------------------------------------------
        everyone                Optional boolean. Default is False, don't share with
                                everyone.
        ----------------------  --------------------------------------------------------
        org                     Optional boolean. Default is False, don't share with
                                the organization.
        ----------------------  --------------------------------------------------------
        groups                  Optional list of group ids as strings, or a list of
                                arcgis.gis.Group objects, or a comma-separated list of
                                group IDs.
        ----------------------  --------------------------------------------------------
        allow_members_to_edit   Optional boolean. Default is False, to allow item to be
                                shared with groups that allow shared update
        ======================  ========================================================

        :return:
            A dictionary with key "notSharedWith" containing array of groups with which the items could not be shared.
        """
        site = self._gis.sites.get(self.site_id)
        result1 = site.item.share(
            everyone=everyone,
            org=org,
            groups=groups,
            allow_members_to_edit=allow_members_to_edit,
        )
        result2 = self.item.share(
            everyone=everyone,
            org=org,
            groups=groups,
            allow_members_to_edit=allow_members_to_edit,
        )
        print(result1)
        return result2

    def unshare(self, groups: list) -> dict:
        """
        Stops sharing of the initiative and its associated site with the specified list of groups.

        ================  =========================================================================================
        **Parameter**      **Description**
        ----------------  -----------------------------------------------------------------------------------------
        groups            Required list of group names as strings, or a list of arcgis.gis.Group objects,
                          or a comma-separated list of group IDs.
        ================  =========================================================================================

        :return:
            Dictionary with key "notUnsharedFrom" containing array of groups from which the items could not be unshared.
        """
        site = self._gis.sites.get(self.site_id)
        result1 = site.item.unshare(groups=groups)
        result2 = self.item.unshare(groups=groups)
        print(result1)
        return result2

    def update(self, initiative_properties=None):
        """Updates the initiative.


        .. note::
            For initiative_properties, pass in arguments for only the properties you want to be updated.
            All other properties will be untouched.  For example, if you want to update only the
            initiative's description, then only provide the description argument in initiative_properties.


        =====================     ====================================================================
        **Parameter**              **Description**
        ---------------------     --------------------------------------------------------------------
        initiative_properties     Required dictionary. See URL below for the keys and values.
        =====================     ====================================================================


        To find the list of applicable options for argument `initiative_properties`, please see the *Item*
        :meth:`~arcgis.gis.Item.update` documentation.

        :return:
           A boolean indicating success (True) or failure (False).

        .. code-block:: python

            USAGE EXAMPLE: Update an initiative successfully

            initiative1 = myHub.initiatives.get('itemId12345')
            initiative1.update(initiative_properties={'description':'Create your own initiative to organize people around a shared goal.'})

            >> True
        """
        if initiative_properties:
            _initiative_data = self.definition
            for key, value in initiative_properties.items():
                _initiative_data[key] = value
                if key == "title":
                    title = value
                    # Fetch Initiative Collaboration group
                    try:
                        _collab_group = self._gis.groups.get(self.collab_group_id)
                        _collab_group.update(title=title + " Core Team")
                    except:
                        pass
                    # Fetch Followers Group
                    try:
                        _followers_group = self._gis.groups.get(self.followers_group_id)
                        _followers_group.update(title=title + " Followers")
                    except:
                        pass
                    # Fetch Content Group
                    _content_group = self._gis.groups.get(self.content_group_id)
                    # Update title for group
                    _content_group.update(title=title + " Content")
            return self.item.update(_initiative_data)


class InitiativeManager(object):
    """
    Helper class for managing initiatives within a Hub. This class is not created by users directly.
    An instance of this class, called 'initiatives', is available as a property of the Hub object. Users
    call methods on this 'initiatives' object to manipulate (add, get, search, etc) initiatives.
    """

    def __init__(self, hub, initiative=None):
        self._hub = hub
        self._gis = self._hub.gis

    def add(self, title, description=None, site=None):
        """
        Adds a new initiative to the Hub.

        =================       ====================================================================
        **Parameter**            **Description**
        -----------------       --------------------------------------------------------------------
        title                   Required string.
        -----------------       --------------------------------------------------------------------
        description             Optional string.
        -----------------       --------------------------------------------------------------------
        site                    Optional Site object.
        =================       ====================================================================

        :return:
           The :class:`~arcgis.apps.hub.initiatives.Initiative` object if successfully added, None if unsuccessful.

        .. code-block:: python

            USAGE EXAMPLE: Add an initiative successfully

            initiative1 = myHub.initiatives.add(title='Vision Zero Analysis')
            initiative1.item
        """

        # Define initiative
        if description is None:
            description = "Create your own initiative by combining existing applications with a custom site."
        _snippet = "Create your own initiative by combining existing applications with a custom site. Use this initiative to form teams around a problem and invite your community to participate."
        _item_dict = {
            "type": "Hub Initiative",
            "snippet": _snippet,
            "typekeywords": "Hub, hubInitiative, OpenData",
            "title": title,
            "description": description,
            "licenseInfo": "CC-BY-SA",
            "culture": "{{culture}}",
            "properties": {},
        }

        # Defining content, collaboration and followers groups
        _content_group_title = title + " Content"
        _content_group_dict = {
            "title": _content_group_title,
            "tags": [
                "Hub Group",
                "Hub Content Group",
                "Hub Site Group",
                "Hub Initiative Group",
            ],
            "access": "public",
        }
        _collab_group_title = title + " Core Team"
        _collab_group_dict = {
            "title": _collab_group_title,
            "tags": [
                "Hub Group",
                "Hub Initiative Group",
                "Hub Site Group",
                "Hub Core Team Group",
                "Hub Team Group",
            ],
            "access": "org",
            "capabilities": "updateitemcontrol",
            "membershipAccess": "collaboration",
            "snippet": "Members of this group can create, edit, and manage the site, pages, and other content related to hub-groups.",
        }
        _followers_group_title = title + " Followers"
        _followers_group_dict = {
            "title": _followers_group_title,
            "tags": [
                "Hub Group",
                "Hub Initiative Group",
                " Hub Initiative Followers Group",
            ],
            "access": "public",
        }

        # Create groups
        content_group = self._gis.groups.create_from_dict(_content_group_dict)
        # Protect groups from accidental deletion
        content_group.protected = True
        # Adding it to _item_dict
        _item_dict["properties"]["contentGroupId"] = content_group.id
        if self._gis.users.me.role == "org_admin":
            collab_group = self._gis.groups.create_from_dict(_collab_group_dict)
            collab_group.protected = True
            _item_dict["properties"]["collaborationGroupId"] = collab_group.id
        if self._hub._hub_enabled:
            followers_group = self._gis.groups.create_from_dict(_followers_group_dict)
            followers_group.protected = True
            _item_dict["properties"]["followersGroupId"] = followers_group.id

        # Create initiative and share it with collaboration group if exists
        item = self._gis.content.add(_item_dict, owner=self._gis.users.me.username)
        try:
            item.share(groups=[collab_group])
        except:
            pass

        # Create initiative site and set initiative properties
        _initiative = Initiative(self._gis, item)
        # If it is a brand new initiative, create new site
        if site is None:
            site = _initiative.sites.add(title=title)
        # else clone existing site
        else:
            site = _initiative.sites.clone(site, pages=True, title=title)
        item.update(
            item_properties={
                "url": site.url,
                "culture": self._gis.properties.user.culture,
            }
        )
        _initiative.site_url = site.item.url
        item.properties["site_id"] = site.itemid

        # update initiative data
        _item_data = {
            "assets": [
                {
                    "id": "bannerImage",
                    "properties": {
                        "type": "resource",
                        "fileName": "detail-image.jpg",
                        "mimeType": "image/jepg",
                    },
                    "license": {"type": "none"},
                    "display": {"position": {"x": "center", "y": "center"}},
                },
                {
                    "id": "iconDark",
                    "properties": {
                        "type": "resource",
                        "fileName": "icon-dark.png",
                        "mimeType": "image/png",
                    },
                    "license": {"type": "none"},
                },
                {
                    "id": "iconLight",
                    "properties": {
                        "type": "resource",
                        "fileName": "icon-light.png",
                        "mimeType": "image/png",
                    },
                    "license": {"type": "none"},
                },
            ],
            "steps": [
                {
                    "id": "informTools",
                    "title": "Inform the Public",
                    "description": "Share data about your initiative with the public so people can easily find, download and use your data in different formats.",
                    "templateIds": [],
                    "itemIds": [site.itemid],
                },
                {
                    "id": "listenTools",
                    "title": "Listen to the Public",
                    "description": "Create ways to gather citizen feedback to help inform your city officials.",
                    "templateIds": [],
                    "itemIds": [],
                },
                {
                    "id": "monitorTools",
                    "title": "Monitor Progress",
                    "description": "Establish performance measures that incorporate the publics perspective.",
                    "templateIds": [],
                    "itemIds": [],
                },
            ],
            "indicators": [],
            "values": {
                "bannerImage": {
                    "source": "bannerImage",
                    "display": {"position": {"x": "center", "y": "center"}},
                },
            },
        }
        _data = json.dumps(_item_data)
        item.update(item_properties={"text": _data})
        return Initiative(self._gis, item)

    def clone(self, initiative, origin_hub=None, title=None):
        """
        Clone allows for the creation of an initiative that is derived from the current initiative.

        .. note::
            If both your `origin_hub` and `destination_hub` are Hub Basic organizations, please use the
            `clone` method supported under the :class:`~arcgis.apps.hub.sites.SiteManager` class.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        initiative          Required :class:`~arcgis.apps.hub.initiatives.Initiative` object of initiative to be cloned.
        ---------------     --------------------------------------------------------------------
        origin_hub          Optional :class:`~arcgis.apps.hub.hub.Hub` object. Required only for cross-org clones where the
                            initiative being cloned is not an item with public access.
        ---------------     --------------------------------------------------------------------
        title               Optional String.
        ===============     ====================================================================

        :return:
            :class:`~arcgis.apps.hub.Initiative` object

        .. code-block:: python

            USAGE EXAMPLE: Clone an initiative in another organization

            #Connect to Hub
            hub_origin = gis1.hub
            hub_destination = gis2.hub
            #Fetch initiative
            initiative1 = hub_origin.initiatives.get('itemid12345')
            #Clone in another Hub
            initiative_cloned = hub_destination.initiatives.clone(initiative1, origin_hub=hub_origin)
            initiative_cloned.item


            USAGE EXAMPLE: Clone initiative in the same organization

            myhub = gis.hub
            initiative1 = myhub.initiatives.get('itemid12345')
            initiative2 = myhub.initiatives.clone(initiative1, title='New Initiative')

        """
        from datetime import timezone

        now = datetime.now(timezone.utc)
        # Checking if item of correct type has been passed
        if "hubInitiative" not in initiative.item.typeKeywords:
            raise Exception("Incorrect item type. Initiative item needed for cloning.")
        # Checking if initiative or site needs to be cloned
        if self._hub and origin_hub:
            if not self._hub._hub_enabled and not origin_hub._hub_enabled:
                raise Exception(
                    "For Hub Basic organizations, please clone the site instead of initiative."
                )
        # New title
        if title is None:
            title = initiative.title + "-copy-%s" % int(now.timestamp() * 1000)
        # If cloning within same org
        if origin_hub is None:
            origin_hub = self._hub
        # Fetch site (checking if origin_hub is correct or if initiative is public)
        try:
            site = origin_hub.sites.get(initiative.site_id)
        except:
            raise Exception(
                "Please provide origin_hub of the initiative object, if the initiative is not publicly shared."
            )
        # Create new initiative if destination hub is premium
        if self._hub._hub_enabled:
            # new initiative
            new_initiative = self._hub.initiatives.add(title=title, site=site)
            return new_initiative
        else:
            # Create new site if destination hub is basic/enterprise
            new_site = self._hub.sites.clone(site, pages=True, title=title)
            return new_site

    def get(self, initiative_id: str) -> str:
        """
        Returns the initiative object for the specified initiative_id.

        =======================    =============================================================
        **Parameter**               **Description**
        -----------------------    -------------------------------------------------------------
        initiative_id              Required string. The initiative itemid.
        =======================    =============================================================

        :return:
            The :class:`~arcgis.apps.hub.initiatives.Initiative` object if the item is found, None if the item is not found.

        .. code-block:: python

            USAGE EXAMPLE: Fetch an initiative successfully

            initiative1 = myHub.initiatives.get('itemId12345')
            initiative1.item

        """
        initiativeItem = self._gis.content.get(initiative_id)
        if "hubInitiative" in initiativeItem.typeKeywords:
            return Initiative(self._gis, initiativeItem)
        else:
            raise TypeError("Item is not a valid initiative or is inaccessible.")

    def search(self, title=None, owner=None, created=None, modified=None, tags=None):
        """
        Searches for initiatives.

        ===============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        title               Optional string. Return initiatives with provided string in title.
        ---------------     --------------------------------------------------------------------
        owner               Optional string. Return initiatives owned by a username.
        ---------------     --------------------------------------------------------------------
        created             Optional string. Date the initiative was created.
                            Shown in milliseconds since UNIX epoch.
        ---------------     --------------------------------------------------------------------
        modified            Optional string. Date the initiative was last modified.
                            Shown in milliseconds since UNIX epoch
        ---------------     --------------------------------------------------------------------
        tags                Optional string. User-defined tags that describe the initiative.
        ===============     ====================================================================

        :return:
           A list of matching :class:`~arcgis.apps.hub.Initiative` objects.
        """

        initiativelist = []

        # Build search query
        query = "typekeywords:hubInitiative"
        if title != None:
            query += " AND title:" + title
        if owner != None:
            query += " AND owner:" + owner
        if created != None:
            query += " AND created:" + created
        if modified != None:
            query += " AND modified:" + modified
        if tags != None:
            query += " AND tags:" + tags

        # Search
        items = self._gis.content.search(query=query, max_items=5000)

        # Return searched initiatives
        for item in items:
            initiativelist.append(Initiative(self._gis, item))
        return initiativelist



from __future__ import annotations
from arcgis.auth.tools._lazy import LazyLoader

from typing import Any
from dataclasses import dataclass, field
from enum import Enum

arcgis = LazyLoader("arcgis")

__all__ = [
    "ItemTypeEnum",
    "ItemProperties",
    "CreateServiceParameter",
    "MetadataFormatEnum",
    "ServiceTypeEnum",
]
###########################################################################


def _parse_enum(value: Enum | Any | None) -> Any | None:
    """returns the Enum's value or the current value"""
    if isinstance(value, Enum):
        return value.value
    else:
        return value


###########################################################################
class ItemTypeEnum(Enum):
    VR_EXPERIENCE = "360 VR Experience"
    CITYENGINE_WEB_SCENE = "CityEngine Web Scene"
    MAP_AREA = "Map Area"
    PRO_MAP = "Pro Map"
    WEB_MAP = "Web Map"
    WEB_SCENE = "Web Scene"
    FEATURE_COLLECTION = "Feature Collection"
    FEATURE_COLLECTION_TEMPLATE = "Feature Collection Template"
    FEATURE_SERVICE = "Feature Service"
    GEODATA_SERVICE = "Geodata Service"
    GLOBE_SERVICE = "Globe Service"
    IMAGE_SERVICE = "Image Service"
    KML = "KML"
    KML_COLLECTION = "KML Collection"
    MAP_SERVICE = "Map Service"
    OGCFEATURESERVER = "OGCFeatureServer"
    ORIENTED_IMAGERY_CATALOG = "Oriented Imagery Catalog"
    RELATIONAL_DATABASE_CONNECTION = "Relational Database Connection"
    SCENE_SERVICE = "Scene Service"
    VECTOR_TILE_SERVICE = "Vector Tile Service"
    WFS = "WFS"
    WMS = "WMS"
    WMTS = "WMTS"
    GEOMETRY_SERVICE = "Geometry Service"
    GEOCODING_SERVICE = "Geocoding Service"
    GEOPROCESSING_SERVICE = "Geoprocessing Service"
    NETWORK_ANALYSIS_SERVICE = "Network Analysis Service"
    WORKFLOW_MANAGER_SERVICE = "Workflow Manager Service"
    APPBUILDER_EXTENSION = "AppBuilder Extension"
    APPBUILDER_WIDGET_PACKAGE = "AppBuilder Widget Package"
    CODE_ATTACHMENT = "Code Attachment"
    DASHBOARD = "Dashboard"
    DEEP_LEARNING_STUDIO_PROJECT = "Deep Learning Studio Project"
    ESRI_CLASSIFICATION_SCHEMA = "Esri Classification Schema"
    EXCALIBUR_IMAGERY_PROJECT = "Excalibur Imagery Project"
    EXPERIENCE_BUILDER_WIDGET = "Experience Builder Widget"
    EXPERIENCE_BUILDER_WIDGET_PACKAGE = "Experience Builder Widget Package"
    FORM = "Form"
    GEOBIM_APPLICATION = "GeoBIM Application"
    GEOBIM_PROJECT = "GeoBIM Project"
    HUB_EVENT = "Hub Event"
    HUB_INITIATIVE = "Hub Initiative"
    HUB_INITIATIVE_TEMPLATE = "Hub Initiative Template"
    HUB_PAGE = "Hub Page"
    HUB_PROJECT = "Hub Project"
    HUB_SITE_APPLICATION = "Hub Site Application"
    INSIGHTS_WORKBOOK = "Insights Workbook"
    INSIGHTS_WORKBOOK_PACKAGE = "Insights Workbook Package"
    INSIGHTS_MODEL = "Insights Model"
    INSIGHTS_PAGE = "Insights Page"
    INSIGHTS_THEME = "Insights Theme"
    INSIGHTS_DATA_ENGINEERING_WORKBOOK = "Insights Data Engineering Workbook"
    INSIGHTS_DATA_ENGINEERING_MODEL = "Insights Data Engineering Model"
    INVESTIGATION = "Investigation"
    MISSION = "Mission"
    MOBILE_APPLICATION = "Mobile Application"
    NOTEBOOK = "Notebook"
    NATIVE_APPLICATION = "Native Application"
    NATIVE_APPLICATION_INSTALLER = "Native Application Installer"
    OPERATION_VIEW = "Operation View"
    OPERATIONS_DASHBOARD_ADD_IN = "Operations Dashboard Add In"
    OPERATIONS_DASHBOARD_EXTENSION = "Operations Dashboard Extension"
    ORTHO_MAPPING_PROJECT = "Ortho Mapping Project"
    ORTHO_MAPPING_TEMPLATE = "Ortho Mapping Template"
    SOLUTION = "Solution"
    STORYMAP = "StoryMap"
    WEB_APPBUILDER_WIDGET = "Web AppBuilder Widget"
    WEB_EXPERIENCE = "Web Experience"
    WEB_EXPERIENCE_TEMPLATE = "Web Experience Template"
    WEB_MAPPING_APPLICATION = "Web Mapping Application"
    WORKFORCE_PROJECT = "Workforce Project"
    ADMINISTRATIVE_REPORT = "Administrative Report"
    APACHE_PARQUET = "Apache Parquet"
    CAD_DRAWING = "CAD Drawing"
    COLOR_SET = "Color Set"
    CONTENT_CATEGORY_SET = "Content Category Set"
    CSV = "CSV"
    DOCUMENT_LINK = "Document Link"
    EARTH_CONFIGURATION = "Earth configuration"
    ESRI_CLASSIFIER_DEFINITION = "Esri Classifier Definition"
    EXPORT_PACKAGE = "Export Package"
    FILE_GEODATABASE = "File Geodatabase"
    GEOJSON = "GeoJson"
    GEOPACKAGE = "GeoPackage"
    GML = "GML"
    IMAGE = "Image"
    IWORK_KEYNOTE = "iWork Keynote"
    IWORK_NUMBERS = "iWork Numbers"
    IWORK_PAGES = "iWork Pages"
    MICROSOFT_EXCEL = "Microsoft Excel"
    MICROSOFT_POWERPOINT = "Microsoft Powerpoint"
    MICROSOFT_WORD = "Microsoft Word"
    PDF = "PDF"
    REPORT_TEMPLATE = "Report Template"
    SERVICE_DEFINITION = "Service Definition"
    SHAPEFILE = "Shapefile"
    SQLITE_GEODATABASE = "SQLite Geodatabase"
    STATISTICAL_DATA_COLLECTION = "Statistical Data Collection"
    STORYMAP_THEME = "StoryMap Theme"
    STYLE = "Style"
    SYMBOL_SET = "Symbol Set"
    VISIO_DOCUMENT = "Visio Document"
    ARCPAD_PACKAGE = "ArcPad Package"
    COMPACT_TILE_PACKAGE = "Compact Tile Package"
    EXPLORER_LAYER = "Explorer Layer"
    IMAGE_COLLECTION = "Image Collection"
    LAYER = "Layer"
    LAYER_PACKAGE = "Layer Package"
    PRO_REPORT = "Pro Report"
    SCENE_PACKAGE = "Scene Package"
    MOBILE_SCENE_PACKAGE = "Mobile Scene Package"
    PROJECT_PACKAGE = "Project Package"
    PROJECT_TEMPLATE = "Project Template"
    PUBLISHED_MAP = "Published Map"
    SCENE_DOCUMENT = "Scene Document"
    TASK_FILE = "Task File"
    TILE_PACKAGE = "Tile Package"
    VECTOR_TILE_PACKAGE = "Vector Tile Package"
    WINDOWS_MOBILE_PACKAGE = "Windows Mobile Package"
    DESKTOP_STYLE = "Desktop Style"
    ARCGIS_PRO_CONFIGURATION = "ArcGIS Pro Configuration"
    DEEP_LEARNING_PACKAGE = "Deep Learning Package"
    GEOPROCESSING_PACKAGE = "Geoprocessing Package"
    GEOPROCESSING_PACKAGE_PRO_VERSION = "Geoprocessing Package (Pro version)"
    GEOPROCESSING_SAMPLE = "Geoprocessing Sample"
    LOCATOR_PACKAGE = "Locator Package"
    RASTER_FUNCTION_TEMPLATE = "Raster function template"
    RULE_PACKAGE = "Rule Package"
    ARCGIS_PRO_ADD_IN = "ArcGIS Pro Add In"
    CODE_SAMPLE = "Code Sample"
    DESKTOP_ADD_IN = "Desktop Add In"
    DESKTOP_APPLICATION = "Desktop Application"
    DESKTOP_APPLICATION_TEMPLATE = "Desktop Application Template"
    EXPLORER_ADD_IN = "Explorer Add In"
    SURVEY123_ADD_IN = "Survey123 Add In"
    WORKFLOW_MANAGER_PACKAGE = "Workflow Manager Package"


###########################################################################
class MetadataFormatEnum(Enum):
    FGDB = "fgdb"
    INSPIRE = "inspire"
    ISO19139 = "iso19139"
    ISO19139_32 = "iso19139-3.2"
    ISO19115 = "iso19115"


###########################################################################
class ServiceTypeEnum(Enum):
    FEATURE_SERVICE = "featureService"
    IMAGE_SERVICE = "imageService"
    RELATIONSHIP_SERVICE = "relationalCatalogService"


###########################################################################
@dataclass
class ItemProperties:
    """
    Item parameters correspond to properties of an item that are available
    to update on the :meth:`~arcgis.gis.ContentManager.add` and
    :meth:`~arcgis.gis.Item.update` operations.
    """

    title: str
    item_type: ItemTypeEnum | str
    tags: list[str] | None = None
    thumbnail: str | None = None
    thumbnail_url: str | None = None
    metadata: str | None = None
    metadata_editable: bool | None = None
    metadata_formats: MetadataFormatEnum | str | None = MetadataFormatEnum.ISO19139
    type_keywords: list[str] | None = None
    description: str | None = None
    snippet: str | None = None
    extent: str | list = None
    spatial_reference: str | None = None
    access_information: str | None = None
    license_info: str | None = None
    culture: str | None = None
    properties: dict | None = None
    app_categories: list[str] | None = None
    industries: list[str] | None = None
    listing_properties: dict | None = None
    service_username: str | None = None
    service_password: str | None = None
    service_proxy: dict | None = None
    categories: list[str] | None = None
    text: dict | str | None = None
    extension: str | None = None
    _dict_data: dict | None = field(init=False)

    def __str__(self):
        return f"<ItemProperties: title={self.title}, type={self.item_type}>"

    def __repr__(self):
        return self.__str__()

    def __post_init__(self):
        self._dict_data = {
            "title": self.title,
            "type": _parse_enum(self.item_type),
            "tags": ",".join(self.tags or []),
            "thumbnail": self.thumbnail,
            "thumbnailurl": self.thumbnail_url,
            "metadata": self.metadata,
            "metadataEditable": self.metadata_editable,
            "metadataFormats": _parse_enum(self.metadata_formats),
            "typeKeywords": ",".join(self.type_keywords or []),
            "description": self.description or "",
            "snippet": self.snippet,
            "extent": self.extent,
            "spatialReference": self.spatial_reference or "",
            "accessInformation": self.access_information,
            "licenseInfo": self.license_info,
            "culture": self.culture,
            "properties": self.properties,
            "appCategories": ",".join(self.app_categories or []),
            "industries": ",".join(self.industries or []),
            "listingProperties": self.listing_properties,
            "serviceUsername": self.service_username,
            "servicePassword": self.service_password,
            "serviceProxyFilter": self.service_proxy,
            "categories": ",".join(self.categories or []),
            "text": self.text or None,
            "extension": self.extension or None,
        }

    def to_dict(self):
        return {
            "title": self.title,
            "type": _parse_enum(self.item_type),
            "tags": ",".join(self.tags or []),
            "thumbnail": self.thumbnail,
            "thumbnailurl": self.thumbnail_url,
            "metadata": self.metadata,
            "metadataEditable": self.metadata_editable,
            "metadataFormats": _parse_enum(self.metadata_formats),
            "typeKeywords": ",".join(self.type_keywords or []),
            "description": self.description or "",
            "snippet": self.snippet,
            "extent": self.extent,
            "spatialReference": self.spatial_reference or "",
            "accessInformation": self.access_information,
            "licenseInfo": self.license_info,
            "culture": self.culture,
            "properties": self.properties,
            "appCategories": ",".join(self.app_categories or []),
            "industries": ",".join(self.industries or []),
            "listingProperties": self.listing_properties,
            "serviceUsername": self.service_username,
            "servicePassword": self.service_password,
            "serviceProxyFilter": self.service_proxy,
            "categories": ",".join(self.categories or []),
            "text": self.text or None,
            "extension": self.extension or None,
        }

    @classmethod
    def fromitem(cls, item: arcgis.gis.Item) -> ItemProperties:
        """Loads the ItemProperties from an `Item`"""
        return ItemProperties(
            title=item.title,
            item_type=ItemTypeEnum._value2member_map_[item.type],
            tags=item.tags,
            type_keywords=item.typeKeywords,
            description=item.description,
            snippet=item.snippet,
            extent=item.extent,
            spatial_reference=item.spatialReference,
            access_information=item.accessInformation,
            license_info=item.licenseInfo,
            culture=item.culture,
            properties=item.properties,
            app_categories=item.appCategories,
            service_proxy=item.proxyFilter,
            industries=item.industries,
            categories=item.categories,
        )


###########################################################################
@dataclass
class CreateServiceParameter:
    """

    The create service parameter description.

    =======================    =============================================================
    **Parameter**               **Description**
    -----------------------    -------------------------------------------------------------
    name                       Required String. Name of the Service
    -----------------------    -------------------------------------------------------------
    output_type                Required :class:`~arcgis.gis._impl._dataclasses.ServiceTypeEnum`
                               or string. The type of service to create.
    -----------------------    -------------------------------------------------------------
    service_description        Optional String. Description given to the service.
    -----------------------    -------------------------------------------------------------
    is_view                    Optional Bool. Specifies if the request is generating a view. The default is False.
    -----------------------    -------------------------------------------------------------
    description                Optional String. A user-friendly description for the published dataset.
    -----------------------    -------------------------------------------------------------
    tags                       Optional String. The tags for the resulting Item
    -----------------------    -------------------------------------------------------------
    snippet                    Optional String. A short description of the Item.
    -----------------------    -------------------------------------------------------------
    static_data                Optional Boolean. Boolean value indicating whether the data changes.
    -----------------------    -------------------------------------------------------------
    max_record_count           Optional Integer. An integer value indicating any constraints enforced on query operations.
    -----------------------    -------------------------------------------------------------
    query_formats              Optional String. The formats in which query results are returned. The default is JSON.
    -----------------------    -------------------------------------------------------------
    capabilities               Optional String. Specify the service capabilities.
    -----------------------    -------------------------------------------------------------
    copyright_text             Optional String. Copyright information associated with the dataset.
    -----------------------    -------------------------------------------------------------
    spatial_reference          Optional Dictionary. All layers added to a hosted feature service need to have the same
                               spatial reference defined for the feature service. When creating a new
                               empty service without specifying its spatial reference, the spatial
                               reference of the hosted feature service is set to the first layer added
                               to that feature service.
    -----------------------    -------------------------------------------------------------
    initial_extent             Optional Dictionary. The initial extent set for the service.
    -----------------------    -------------------------------------------------------------
    allow_geometry_updates     Optional Boolean. Boolean value indicating if updating the geometry of the service is permitted.
    -----------------------    -------------------------------------------------------------
    units                      Optional String. Units used by the feature service.
    -----------------------    -------------------------------------------------------------
    xssPreventionInfo          Optional Dictionary. A JSON object specifying the properties of cross-site scripting prevention.
    -----------------------    -------------------------------------------------------------
    overwrite                  Optional Boolean. Overwrite an existing service when true. The default is False
    -----------------------    -------------------------------------------------------------
    itemid_to_create           Optional String. The item id to create.  This is only valid on ArcGIS Enterprise.
    =======================    =============================================================
    """

    name: str
    output_type: ServiceTypeEnum | str = ServiceTypeEnum.FEATURE_SERVICE
    itemid_to_create: str | None = None
    overwrite: bool = False
    service_description: str | None = ""
    is_view: bool = False
    description: str | None = None
    tags: list[str] | None = None
    snippet: str | None = None
    static_data: bool = True
    max_record_count: int = 1000
    query_formats: str = "JSON"
    capabilities: str = "Query"
    copyright_text: str | None = None
    spatial_reference: dict | None = None
    initial_extent: dict | None = None
    allow_geometry_updates: bool | None = None
    units: str | None = None
    xssPreventionInfo: dict | None = None
    provider: str | None = None
    connection: dict | None = None
    _dict_data: dict | None = field(init=False)

    def __str__(self):
        return "<CreateServiceParameter>"

    def __repr__(self):
        return "<CreateServiceParameter>"

    def __post_init__(self):
        self._dict_data = self._create_dict()

    def _create_dict(self):
        default_sr = {"wkid": 102100}
        init_extent = {
            "xmin": -20037507.0671618,
            "ymin": -30240971.9583862,
            "xmax": 20037507.0671618,
            "ymax": 18398924.324645,
            "spatialReference": {"wkid": 102100, "latestWkid": 3857},
        }
        init_xss = {
            "xssPreventionEnabled": True,
            "xssPreventionRule": "input",
            "xssInputRule": "rejectInvalid",
        }
        if (
            _parse_enum(self.output_type) == "imageService"
            and self.capabilities is None
        ):
            self.capabilities = "Image,Catalog,Mensuration"
        self._dict_data = {
            "outputType": _parse_enum(self.output_type),
            "isView": self.is_view,
            "description": self.description or "",
            "snippet": self.snippet or "",
            "tags": ",".join(self.tags or []),
            "itemIdToCreate": self.itemid_to_create or None,
            "createParameters": {
                "name": self.name,
                "serviceDescription": self.service_description,
                "hasStaticData": self.static_data,
                "maxRecordCount": self.max_record_count,
                "supportedQueryFormats": self.query_formats or "JSON",
                "capabilities": self.capabilities or "Query",
                "copyRightText": self.copyright_text or "",
                "spatialReference": self.spatial_reference or default_sr,
                "initialExtent": self.initial_extent or init_extent,
                "allowGeometryUpdates": self.allow_geometry_updates,
                "units": self.units or "esriMeters",
                "xssPreventionInfo": self.xssPreventionInfo or init_xss,
            },
        }
        if _parse_enum(self.output_type) == "relationalCatalogService":
            self._dict_data["createParameters"]["provider"] = "ADS"
            self._dict_data["createParameters"][
                "connectionProperties"
            ] = self.connection

        if self._dict_data["itemIdToCreate"] in [None, ""]:
            del self._dict_data["itemIdToCreate"]
        return self._dict_data

    def to_dict(self) -> dict:
        return self._create_dict()



import numpy as np


def multi_cat_reweighing(df, _sensitive_feature, _target):
    df["weight"] = 1

    total_count = df["weight"].sum()

    total_no_of_favours = np.count_nonzero(df[_target] == 1)
    total_no_of_unfavours = np.count_nonzero(df[_target] == 0)

    categories = df[_sensitive_feature].unique()
    weights = []

    for val in categories:
        total_no_of_cat = np.count_nonzero(df[_sensitive_feature] == val)
        total_no_of_cat_with_favour = df[
            (df[_sensitive_feature] == val) & (df[_target] == 1)
        ][_target].count()
        total_no_of_cat_with_unfavour = df[
            (df[_sensitive_feature] == val) & (df[_target] == 0)
        ][_target].count()

        w_of_cat_with_favour = (
            total_no_of_favours
            * total_no_of_cat
            / (total_count * total_no_of_cat_with_favour)
        )
        w_of_cat_with_unfavour = (
            total_no_of_unfavours
            * total_no_of_cat
            / (total_count * total_no_of_cat_with_unfavour)
        )

        weights.append((val, w_of_cat_with_favour, w_of_cat_with_unfavour))

    return weights


def assign_weights(_df, _target, _sensitive_feature):
    weights = multi_cat_reweighing(_df, _sensitive_feature, _target)
    for c, w_fav, w_unfav in weights:
        cond_p_fav = np.where((_df[_sensitive_feature] == c) & (_df[_target] == 1))[0]
        _df.iloc[cond_p_fav, -1] = w_fav

        cond_p_unfav = np.where((_df[_sensitive_feature] == c) & (_df[_target] == 0))[0]
        _df.iloc[cond_p_unfav, -1] = w_unfav

    return _df["weight"]



from arcgis.widgets._mapview._raster.engine.matplotlib_raster_engine import (
    MatplotlibRasterEngine,
)
from arcgis.widgets._mapview._raster.engine.arcpy_raster_engine import ArcPyRasterEngine
from arcgis.widgets._mapview._raster.engine.rasterio_engine import RasterIOEngine



from __future__ import print_function
import datetime

import inspect
import logging
import sys
import json
import time
import datetime
import collections
import concurrent.futures

import arcgis
from arcgis.gis import GIS
from arcgis.gis._impl._con import Connection
from arcgis.features import FeatureSet, FeatureCollection, Table
from arcgis.auth.tools import LazyLoader

mapping = LazyLoader("arcgis.mapping")
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing._tool import _camelCase_to_underscore
from arcgis._impl.common._utils import _date_handler
from arcgis.geoprocessing._job import GPJob

_log = logging.getLogger(__name__)


def _layer_input(input_layer):
    # Will be used exclusively by RA tools
    input_param = input_layer

    url = ""
    if isinstance(input_layer, arcgis.gis.Item):
        if input_layer.type == "Image Collection":
            input_param = {"itemId": input_layer.itemid}
        else:
            if "layers" in input_layer:
                input_param = input_layer.layers[0]._lyr_dict
            else:
                raise TypeError("No layers in input layer Item")

    elif isinstance(input_layer, arcgis.features.FeatureLayerCollection):
        input_param = input_layer.layers[0]._lyr_dict

    elif isinstance(input_layer, arcgis.features.FeatureCollection):
        input_param = input_layer.properties

    elif isinstance(input_layer, arcgis.gis.Layer):
        input_param = input_layer._lyr_dict
        from arcgis.raster import ImageryLayer
        import json

        if isinstance(input_layer, ImageryLayer):
            if "options" in input_layer._lyr_json:
                if isinstance(
                    input_layer._lyr_json["options"], str
                ):  # sometimes the rendering info is a string
                    # load json
                    layer_options = json.loads(input_layer._lyr_json["options"])
                else:
                    layer_options = input_layer._lyr_json["options"]

                if "imageServiceParameters" in layer_options:
                    # get renderingRule and mosaicRule
                    input_param.update(layer_options["imageServiceParameters"])

    elif isinstance(input_layer, dict):
        input_param = input_layer

    elif isinstance(input_layer, str):
        if "http:" in input_layer or "https:" in input_layer:
            input_param = {"url": input_layer}
        else:
            input_param = {"uri": input_layer}

    else:
        raise Exception(
            "Invalid format of input layer. url string, layer Item, layer instance or dict supported"
        )

    if "url" in input_param:
        url = input_param["url"]
        if "/RasterRendering/" in url:
            url = input_layer._uri
            input_param = {"uri": url}
            return input_param
    if "ImageServer" in url or "MapServer" in url:
        if "serviceToken" in input_param:
            url = url + "?token=" + input_param["serviceToken"]
            input_param.update({"url": url})

    return input_param


def _layer_input_gp(input_layer):
    input_param = input_layer

    input_layer_url = ""
    if isinstance(input_layer, arcgis.gis.Item):
        if "layers" in input_layer:
            input_param = input_layer.layers[0]._lyr_dict
        else:
            raise TypeError("No layers in input layer Item")

    elif isinstance(input_layer, arcgis.features.FeatureLayerCollection):
        input_param = input_layer.layers[0]._lyr_dict

    elif isinstance(input_layer, arcgis.features.FeatureCollection):
        input_param = input_layer.properties

    elif isinstance(input_layer, arcgis.features.FeatureSet):
        input_param = input_layer.to_dict()

    elif isinstance(input_layer, arcgis.gis.Layer):
        input_param = input_layer._lyr_dict

    elif isinstance(input_layer, dict):
        input_param = input_layer

    elif isinstance(input_layer, str):
        if "http:" in input_layer or "https:" in input_layer:
            input_param = {"url": input_layer}
        else:
            input_param = {"uri": input_layer}

    else:
        raise Exception(
            "Invalid format of input layer. url string, layer Item, layer instance or dict supported"
        )

    return input_param


def _feature_input(input_layer):
    input_param = input_layer

    input_layer_url = ""
    if isinstance(input_layer, arcgis.gis.Item):
        if input_layer.type.lower() == "feature service":
            input_param = input_layer.layers[0]._lyr_dict
        elif input_layer.type.lower() == "big data file share":
            input_param = input_layer.layers[0]._lyr_dict
        elif input_layer.type.lower() == "feature collection":
            fcdict = input_layer.get_data()
            fc = FeatureCollection(fcdict["layers"][0])
            input_param = fc.layer
        else:
            raise TypeError("item type must be feature service or feature collection")

    elif isinstance(input_layer, arcgis.features.FeatureLayerCollection):
        input_param = input_layer.layers[0]._lyr_dict

    elif isinstance(input_layer, arcgis.features.FeatureCollection):
        input_param = input_layer.properties

    elif isinstance(input_layer, arcgis.gis.Layer):
        input_param = input_layer._lyr_dict

    elif isinstance(input_layer, dict):
        input_param = input_layer

    elif isinstance(input_layer, str):
        input_param = {"url": input_layer}

    else:
        raise Exception(
            "Invalid format of input layer. url string, feature service Item, feature service instance or dict supported"
        )

    return input_param


def _analysis_job(gptool, task, params):
    """Submits an Analysis job and returns the job URL for monitoring the job
    status in addition to the json response data for the submitted job."""

    # Unpack the Analysis job parameters as a dictionary and add token and
    # formatting parameters to the dictionary. The dictionary is used in the
    # HTTP POST request. Headers are also added as a dictionary to be included
    # with the POST.
    #
    # print("Submitting analysis job...")

    task_url = "{}/{}".format(gptool.url, task)
    submit_url = "{}/submitJob".format(task_url)

    params["f"] = "json"
    try:
        resp = gptool._con.post(submit_url, params, token=gptool._token)
    except RuntimeError:
        resp = gptool._con.post(submit_url, params)
    # print(resp)
    return task_url, resp, resp["jobId"]


def _analysis_job_status(gptool, task_url, job_info):
    """Tracks the status of the submitted Analysis job."""

    if "jobId" in job_info:
        # Get the id of the Analysis job to track the status.
        #
        try:
            job_id = job_info.get("jobId")
            job_url = "{}/jobs/{}".format(task_url, job_id)
            params = {"f": "json"}
            try:
                job_response = gptool._con.get(job_url, params, token=gptool._token)
            except Exception as e:
                job_response = gptool._con.get(job_url, params)

            # Query and report the Analysis job status.
            #
            num_messages = 0
            if "jobStatus" in job_response:
                while not job_response.get("jobStatus") == "esriJobSucceeded":
                    time.sleep(1)
                    try:
                        job_response = gptool._con.get(
                            job_url, params, token=gptool._token
                        )
                    except Exception as e:
                        job_response = gptool._con.get(job_url, params)

                    # print(job_response)
                    messages = (
                        job_response["messages"] if "messages" in job_response else []
                    )
                    num = len(messages)
                    if num > num_messages:
                        for index in range(num_messages, num):
                            msg = messages[index]
                            if arcgis.env.verbose:
                                print(msg["description"])
                            if msg["type"] == "esriJobMessageTypeInformative":
                                _log.info(msg["description"])
                            elif msg["type"] == "esriJobMessageTypeWarning":
                                _log.warning(msg["description"])
                            elif msg["type"] == "esriJobMessageTypeError":
                                _log.error(msg["description"])
                                # print(msg['description'], file=sys.stderr)
                            else:
                                _log.warning(msg["description"])
                        num_messages = num

                    if job_response.get("jobStatus") == "esriJobFailed":
                        raise Exception("Job failed.")
                    elif job_response.get("jobStatus") == "esriJobCancelled":
                        raise Exception("Job cancelled.")
                    elif job_response.get("jobStatus") == "esriJobTimedOut":
                        raise Exception("Job timed out.")

                if "results" in job_response:
                    return job_response
                else:
                    retry_counter = 0
                    while retry_counter < 5:
                        time.sleep(retry_counter + 1)
                        try:
                            job_response = gptool._con.get(
                                job_url, params, token=gptool._token
                            )
                        except Exception as e:
                            job_response = gptool._con.get(job_url, params)
                        if "results" in job_response:
                            return job_response
                        retry_counter += 1
                if "results" in job_response:
                    return job_response
                else:
                    raise Exception("No job results.")

            else:
                raise Exception("No job results.")
        except KeyboardInterrupt:
            cancel_url = "%s/jobs/%s/cancel" % (task_url, job_info["jobId"])
            params = {"f": "json"}
            job_info = gptool._con.get(path=cancel_url, params=params)
            job_info = _analysis_job_status(gptool, task_url, job_info)
    else:
        raise Exception("No job url.")


def _analysis_job_results(gptool, task_url, job_info, job_id=None):
    """Use the job result json to get information about the feature service
    created from the Analysis job."""

    # Get the paramUrl to get information about the Analysis job results.
    #
    if job_id is None:
        job_id = job_info.get("jobId")

    if "results" in job_info:
        results = job_info.get("results")
        result_values = {}
        for key in list(results.keys()):
            param_value = results[key]
            if "paramUrl" in param_value:
                param_url = param_value.get("paramUrl")
                result_url = "{}/jobs/{}/{}".format(task_url, job_id, param_url)

                params = {"f": "json"}
                _set_env_params(params, {})
                try:
                    param_result = gptool._con.get(
                        result_url, params, token=gptool._token
                    )
                except:
                    param_result = gptool._con.get(result_url, params)
                if isinstance(param_result, list):
                    result_values[key] = [value.get("value") for value in param_result]
                else:
                    job_value = param_result.get("value")
                    result_values[key] = job_value
        return result_values
    else:
        raise Exception("Unable to get analysis job results.")


def _future_op(
    gptool,
    task_url,
    job_info,
    job_id,
    param_db,
    return_values,
    return_messages,
):
    job_info = _analysis_job_status(gptool, task_url, job_info)
    resp = _analysis_job_results(gptool, task_url, job_info, job_id)

    # ---------------------async-out---------------------#
    output_dict = {}
    for retParamName in resp.keys():
        output_val = resp[retParamName]
        try:
            ret_param_name, ret_val = _get_output_value(
                gptool, output_val, param_db, retParamName
            )
            output_dict[ret_param_name] = ret_val
        except KeyError:
            pass  # cannot handle unexpected output as return tuple will change

    # tools with output map service - add another output:
    # result_layer = '' #***self.properties.resultMapServerName
    if gptool.properties.resultMapServerName != "":
        job_id = job_info.get("jobId")
        result_layer_url = (
            gptool._url.replace("/GPServer", "/MapServer") + "/jobs/" + job_id
        )

        output_dict["result_layer"] = mapping.MapImageLayer(
            result_layer_url, gptool._gis
        )

    num_returns = len(resp)
    if return_messages:
        return (
            _return_output(num_returns, output_dict, return_values),
            job_info,
        )

    return _return_output(num_returns, output_dict, return_values)


def _execute_gp_tool(
    gis,
    task_name,
    params,
    param_db,
    return_values,
    use_async,
    url,
    webtool=False,
    add_token=True,
    return_messages=False,
    future=False,
):
    if gis is None:
        gis = arcgis.env.active_gis
    elif (  # Checks if the GIS is not a GIS class but has the _con property
        isinstance(gis, GIS) == False
        and hasattr(gis, "_con")
        and isinstance(gis._con, Connection)
    ):
        gis = gis._con

    if isinstance(gis, Connection):
        log = logging.getLogger()
        log.warning("Using Connection object over GIS object")
        ngis = GIS(set_active=False)
        ngis._con = gis
        gis = ngis

    gp_params = {"f": "json"}

    # ---------------------in---------------------#
    for param_name, param_value in params.items():
        # print(param_name + " = " + str(param_value))
        if param_name in param_db:
            py_type, gp_param_name = param_db[param_name]
            if param_value is None:
                param_value = ""
            gp_params[gp_param_name] = param_value
            if py_type == FeatureSet:
                if webtool:
                    if isinstance(param_value, (tuple, list)):
                        gp_params[gp_param_name] = [
                            _layer_input_gp(p) for p in param_value
                        ]
                    else:
                        gp_params[gp_param_name] = _layer_input_gp(param_value)

                else:
                    try:
                        from arcgis.features.geo._accessor import (
                            _is_geoenabled,
                        )
                    except:

                        def _is_geoenabled(o):
                            return False

                    if type(param_value) == FeatureSet:
                        gp_params[gp_param_name] = param_value.to_dict()
                    elif _is_geoenabled(param_value):
                        gp_params[gp_param_name] = json.loads(
                            json.dumps(
                                param_value.spatial.__feature_set__,
                                default=_date_handler,
                            )
                        )
                    elif isinstance(param_value, arcgis.gis.Layer):
                        gp_params[gp_param_name] = _layer_input_gp(param_value)
                    elif type(param_value) == str:
                        try:
                            klass = py_type
                            gp_params[gp_param_name] = klass.from_str(param_value)

                        except:
                            pass

            elif py_type in [LinearUnit, DataFile, RasterData]:
                if type(param_value) in [LinearUnit, DataFile, RasterData]:
                    gp_params[gp_param_name] = param_value.to_dict()

                elif type(param_value) == str:
                    try:
                        klass = py_type
                        gp_params[gp_param_name] = klass.from_str(param_value)

                    except:
                        pass

                elif isinstance(param_value, arcgis.gis.Layer):
                    gp_params[gp_param_name] = param_value._lyr_dict

            elif py_type == datetime.datetime:
                gp_params[gp_param_name] = _date_handler(param_value)
    # --------------------------------------------#

    _set_env_params(gp_params, params)

    # for param_name, param_value in gp_params.items():
    #     print(param_name + " = " + str(param_value))

    gptool = arcgis.gis._GISResource(url, gis)

    if use_async:
        task_url = "{}/{}".format(url, task_name)
        submit_url = "{}/submitJob".format(task_url)
        if add_token and submit_url.lower().find("arcgis.com") == -1:
            try:
                job_info = gptool._con.post(submit_url, gp_params, token=gptool._token)
            except:
                job_info = gptool._con.post(submit_url, gp_params)
        else:
            job_info = gptool._con.post(submit_url, gp_params)
        job_id = job_info["jobId"]
        if future:
            executor = concurrent.futures.ThreadPoolExecutor(1)
            future = executor.submit(
                _future_op,
                *(
                    gptool,
                    task_url,
                    job_info,
                    job_id,
                    param_db,
                    return_values,
                    return_messages,
                ),
            )
            executor.shutdown(False)
            gpjob = GPJob(
                future=future,
                gptool=gptool,
                jobid=job_id,
                task_url=task_url,
                gis=gptool._gis,
                notify=arcgis.env.verbose,
            )
            return gpjob
        job_info = _analysis_job_status(gptool, task_url, job_info)
        resp = _analysis_job_results(gptool, task_url, job_info, job_id)

        # ---------------------async-out---------------------#
        output_dict = {}
        for retParamName in resp.keys():
            output_val = resp[retParamName]
            try:
                ret_param_name, ret_val = _get_output_value(
                    gptool, output_val, param_db, retParamName
                )
                output_dict[ret_param_name] = ret_val
            except KeyError:
                pass  # cannot handle unexpected output as return tuple will change

        # tools with output map service - add another output:
        # result_layer = '' #***self.properties.resultMapServerName
        if gptool.properties.resultMapServerName != "":
            job_id = job_info.get("jobId")
            result_layer_url = (
                url.replace("/GPServer", "/MapServer") + "/jobs/" + job_id
            )

            output_dict["result_layer"] = mapping.MapImageLayer(
                result_layer_url, gptool._gis
            )

        num_returns = len(resp)
        if return_messages:
            return (
                _return_output(num_returns, output_dict, return_values),
                job_info,
            )

        return _return_output(num_returns, output_dict, return_values)

    else:  # synchronous
        exec_url = url + "/" + task_name + "/execute"
        if add_token:
            try:
                resp = gptool._con.post(exec_url, gp_params, token=gptool._token)
            except:
                resp = gptool._con.post(exec_url, gp_params)
        else:
            resp = gptool._con.post(exec_url, gp_params)

        output_dict = {}

        for result in resp["results"]:
            retParamName = result["paramName"]

            output_val = result["value"]
            try:
                ret_param_name, ret_val = _get_output_value(
                    gptool, output_val, param_db, retParamName
                )
                output_dict[ret_param_name] = ret_val
            except KeyError:
                pass  # cannot handle unexpected output as return tuple will change

        num_returns = len(resp["results"])
        if return_messages:
            return (
                _return_output(num_returns, output_dict, return_values),
                job_info,
            )
        return _return_output(num_returns, output_dict, return_values)


def _set_env_params(gp_params, params):
    # copy environment variables if set
    if "env:outSR" not in params and arcgis.env.out_spatial_reference is not None:
        gp_params["env:outSR"] = arcgis.env.out_spatial_reference
    if (
        "env:processSR" not in params
        and arcgis.env.process_spatial_reference is not None
    ):
        gp_params["env:processSR"] = arcgis.env.process_spatial_reference
    if "returnZ" not in params and arcgis.env.return_z is not False:
        gp_params["returnZ"] = True
    if "returnM" not in params and arcgis.env.return_m is not False:
        gp_params["returnM"] = True


def _return_output(num_returns, output_dict, return_values):
    if num_returns == 1:
        return output_dict[return_values[0]["name"]]
    else:
        ret_names = []
        for return_value in return_values:
            ret_names.append(return_value["name"])
        # ret_names = output_dict.keys() # CANT USE - the order matters
        NamedTuple = collections.namedtuple("ToolOutput", ret_names)
        tool_output = NamedTuple(**output_dict)
        return tool_output


def _get_output_value(gptool, output_val, param_db, retParamName):
    ret_param_name = _camelCase_to_underscore(retParamName)

    ret_type, _ = param_db[ret_param_name]

    ret_val = output_val
    if output_val is not None:
        if ret_type in [FeatureSet, LinearUnit, DataFile, RasterData, Table]:
            jsondict = output_val
            if (
                "mapImage" in jsondict
            ):  # http://resources.esri.com/help/9.3/arcgisserver/apis/rest/gpresult.html#mapimage
                ret_val = jsondict
            elif ret_type == Table and "url" in jsondict:
                ret_val = arcgis.features.Table(jsondict["url"], gptool._gis)
            elif ret_type == FeatureSet and "url" in jsondict:
                ret_val = arcgis.features.FeatureLayer(jsondict["url"], gptool._gis)
            elif len(jsondict) == 0 or jsondict == {}:
                ret_val = None
            else:
                result = ret_type.from_dict(jsondict)
                result._con = gptool._con
                result._token = gptool._token
                ret_val = result
        else:
            ret_val = output_val
    return ret_param_name, ret_val



import os

os.environ["USE_TF"] = "0"
os.environ["USE_TORCH"] = "1"
import re
import traceback
import collections
from functools import partial

HAS_TRANSFORMER = True

try:
    import torch
    import pandas as pd
    from fastai.data_block import DataBunch
    from fastai.basic_data import DatasetType
    from torch.utils.data import Dataset, DataLoader
    from transformers import PreTrainedTokenizer, AutoTokenizer
    from fastai.text import (
        List,
        BaseTokenizer,
        Vocab,
        Collection,
        SortishSampler,
        SortSampler,
    )
    from fastai.torch_core import to_data, BatchSamples, Tuple, LongTensor
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False

    class BaseTokenizer:
        pass

    class Vocab:
        pass


def process_text(text):
    text = re.sub(" +", " ", text)
    text = re.sub(" ([@.#$/:-]) ?", r"\1", text)
    # Reomve leading and trailing special characters
    while text and text[0] in [".", ",", "-", ":", "@", "$"]:
        text = text[1:]
    while text and text[-1] in [".", ",", "-", ":", "@", "$"]:
        text = text[:-1]
    return text.strip()


models_subwords_map = {
    "distilbert": ("##", False),
    "bert": ("##", False),
    "electra": ("##", False),
    "xlnet": ("▁", True),
    "camembert": ("▁", True),
    "albert": ("▁", True),
    "xlm-roberta": ("▁", True),
    "roberta": ("Ġ", True),
    "longformer": ("Ġ", True),
    "bart": ("Ġ", True),
    "mobilebert": ("##", False),
    "funnel": ("##", False),
}


class TransformersBaseTokenizer(BaseTokenizer):
    """
    Wrapper around PreTrainedTokenizer to be compatible with fast.ai
    """

    def __init__(
        self,
        pretrained_tokenizer: PreTrainedTokenizer,
        lang="en",
        seq_len=512,
        **kwargs,
    ):
        super().__init__(lang)
        self._pretrained_tokenizer = pretrained_tokenizer
        self._seq_len = seq_len
        self.max_seq_len = min(pretrained_tokenizer.model_max_length, self._seq_len)

    def __call__(self, *args, **kwargs):
        return self

    # new improved tokenizer which can support any transformer model
    def tokenizer(self, t: str) -> List[str]:
        ids = self._pretrained_tokenizer.encode(
            t, max_length=self.max_seq_len, truncation=True
        )
        tokens = self._pretrained_tokenizer.convert_ids_to_tokens(ids)
        return tokens


class TransformersVocab(Vocab):
    """
    Contain the correspondence between numbers and tokens and numericalize.
    """

    def __init__(self, tokenizer: PreTrainedTokenizer):
        super().__init__(itos=[])
        self.tokenizer = tokenizer
        self.special_token_map = self.tokenizer.special_tokens_map
        self.has_cls_token = True if self.special_token_map.get("cls_token") else False
        self.has_sep_token = True if self.special_token_map.get("sep_token") else False
        self.has_unk_token = True if self.special_token_map.get("unk_token") else False
        self.has_pad_token = True if self.special_token_map.get("pad_token") else False

    def numericalize(self, t: Collection[str]) -> List[int]:
        """
        Convert a list of tokens `t` to their ids.
        """
        return self.tokenizer.convert_tokens_to_ids(t)

    def textify(self, nums: Collection[int], sep=" ") -> str:
        """
        Convert a list of `nums` to their tokens.
        """
        # text = self.tokenizer.decode(nums, skip_special_tokens=True)
        text = self.tokenizer.decode(nums)
        if self.has_cls_token:
            text = text.replace(self.tokenizer.cls_token, "")
        if self.has_sep_token:
            text = text.replace(self.tokenizer.sep_token, "")
        if self.has_unk_token:
            text = text.replace(self.tokenizer.unk_token, "")
        if self.has_pad_token:
            text = text.replace(self.tokenizer.pad_token, "")
        text = re.sub(" +", " ", text)
        return text.strip()

    def __getstate__(self):
        return {"itos": self.itos, "tokenizer": self.tokenizer}

    def __setstate__(self, state: dict):
        self.itos = state["itos"]
        self.tokenizer = state["tokenizer"]
        self.stoi = collections.defaultdict(
            int, {v: k for k, v in enumerate(self.itos)}
        )


class TransformerNERDataset(Dataset):
    def __init__(
        self,
        tokens_collection,
        tags_collection,
        tokenizer,
        label2id,
        model_type="bert",
        seq_length=512,
    ):
        self.tokens_collection = tokens_collection
        self.tags_collection = tags_collection
        self.tokenizer = tokenizer
        self.label2id = label2id
        self.id2label = {y: x for x, y in self.label2id.items()}
        self.model_type = model_type
        self.ignore_token_list = ["\n"]
        self.otag_id = (
            self.label2id["O"] if "O" in self.label2id else self.label2id["o"]
        )
        self.max_seq_len = min(tokenizer.model_max_length, seq_length)

    def __len__(self):
        return len(self.tokens_collection)

    def __getitem__(self, index):
        tokens = self.tokens_collection[index]
        tags = self.tags_collection[index]
        ids, target_tags = [], []

        tokens = [x for x in tokens if x not in self.ignore_token_list]
        if self.model_type in ["roberta", "bart", "longformer"]:
            # tokens = [f" {x}" if index != 0 else x for index, x in enumerate(tokens)]
            tokens = [f" {x}" for index, x in enumerate(tokens)]

        for index, token in enumerate(tokens):
            # Transformer tokenizer can split the input word at sub-word level.
            # For example - `uninvited` can be broken into: 'un', '##in', '##vi', '##ted'
            token_breakup_list = self.tokenizer.encode(token, add_special_tokens=False)
            num_items = len(token_breakup_list)
            ids.extend(token_breakup_list)
            tag_id = self.label2id[tags[index]]
            target_tags.extend([tag_id] * num_items)

        # self.max_seq_len - 2, because we will be adding a [CLS] & [SEP] token when creating the DataBunch
        ids = ids[: self.max_seq_len - 2]
        target_tags = target_tags[: self.max_seq_len - 2]

        if self.model_type in ["xlnet"]:
            ids = ids + [self.tokenizer.sep_token_id, self.tokenizer.cls_token_id]
            target_tags = target_tags + [self.otag_id, self.otag_id]
        elif self.model_type in ["xlm", "flaubert"]:
            ids = [self.tokenizer.bos_token_id] + ids + [self.tokenizer.sep_token_id]
            target_tags = [self.otag_id] + target_tags + [self.otag_id]
        else:
            ids = [self.tokenizer.cls_token_id] + ids + [self.tokenizer.sep_token_id]
            target_tags = [self.otag_id] + target_tags + [self.otag_id]

        return {"ids": ids, "target": target_tags, "otag_id": self.otag_id}


def collate_func(
    samples: BatchSamples, tokenizer: PreTrainedTokenizer, seq_len: int = 512
):
    # Function that collect samples and adds padding
    samples = to_data(samples)
    batch_sentences = [x["ids"] for x in samples]
    batch_target = [x["target"] for x in samples]
    otag_id = samples[0]["otag_id"]

    # [CLS] & [SEP] token are already added from the items returned from TransformerNERDataset
    max_len_batch = max([len(x) for x in batch_target])
    max_seq_len = min(tokenizer.model_max_length, seq_len)
    max_len = min(max_len_batch, max_seq_len)

    res_x = tokenizer.batch_encode_plus(
        batch_sentences,
        max_length=max_len,
        padding=True,
        truncation=True,
        is_split_into_words=True,
        add_special_tokens=False,
        return_tensors="pt",
    )

    input_ids, attention_masks = res_x.get("input_ids"), res_x.get("attention_mask")
    token_type_ids = res_x.get("token_type_ids")

    res_y = torch.zeros(len(samples), max_len).long() + otag_id

    pad_first = True if tokenizer.padding_side == "left" else False
    for i, s in enumerate(batch_target):
        if pad_first:
            res_y[i, -len(s) :] = LongTensor(s)
        else:
            res_y[i, : len(s) :] = LongTensor(s)

    # res_y is added in the list as the model has the ability to return the loss if the labels are provided
    if token_type_ids is not None:
        return (
            [input_ids, attention_masks, token_type_ids, res_y],
            [
                res_y,
                attention_masks,
            ],
        )
    else:
        return [input_ids, attention_masks, res_y], [res_y, attention_masks]


class TransformerNERDataBunch(DataBunch):
    "Create a `TextDataBunch` suitable for training an RNN classifier."

    @classmethod
    def create(
        cls,
        train_ds,
        valid_ds,
        tokenizer: PreTrainedTokenizer,
        bs: int = 4,
        val_bs: int = None,
        device: torch.device = None,
        no_check: bool = False,
        seq_len: int = 512,
        **dl_kwargs,
    ) -> DataBunch:
        "Function that transform the `datasets` in a `DataBunch`. Passes `**dl_kwargs` on to `DataLoader()`"
        if val_bs is None:
            val_bs = bs
        # write own collate function
        collate_fn = partial(collate_func, tokenizer=tokenizer, seq_len=seq_len)
        train_sampler = SortishSampler(
            train_ds, key=lambda t: len(train_ds[t]["target"]), bs=bs
        )
        train_dl = DataLoader(
            train_ds, batch_size=bs, sampler=train_sampler, drop_last=False, **dl_kwargs
        )
        dataloaders = [train_dl]

        sampler = SortSampler(valid_ds, key=lambda t: len(valid_ds[t]["target"]))
        dataloaders.append(
            DataLoader(
                valid_ds,
                batch_size=val_bs,
                sampler=sampler,
                drop_last=False,
                **dl_kwargs,
            )
        )
        return cls(
            *dataloaders, device=device, collate_fn=collate_fn, no_check=no_check
        )

    @property
    def is_empty(self) -> bool:
        return not (
            (self.train_dl and len(self.train_ds) != 0)
            or (self.valid_dl and len(self.valid_ds) != 0)
            or (self.test_dl and len(self.test_ds) != 0)
        )

    def show_batch(
        self,
        rows: int = 5,
        ds_type: DatasetType = DatasetType.Train,
        reverse: bool = False,
        **kwargs,
    ):
        x, y = self.one_batch(ds_type, True, True)
        batch_tokens = x[0].tolist()
        batch_labels = y[0].tolist()
        num_items = min(rows, self.dl(DatasetType.Train).batch_size)
        results = get_results(
            batch_tokens,
            batch_labels,
            self.train_ds.tokenizer,
            self.train_ds.id2label,
            self.train_ds.model_type,
            num_items,
        )
        df = pd.DataFrame(
            results,
        )
        df.fillna("", inplace=True)
        return df


def get_previous_tokens(token_ids, index, model_type, tok):
    res = models_subwords_map.get(model_type)
    if res is None:
        return [token_ids[index]]

    i, whole_word_ids = index, []
    spl_char, start_token_contain_spl_char = res
    while i > 0:
        curr_token = tok.convert_ids_to_tokens(token_ids[i])
        whole_word_ids.insert(0, token_ids[i])
        if curr_token.startswith(spl_char):
            if start_token_contain_spl_char:
                break
        else:
            if not start_token_contain_spl_char:
                break
        i -= 1
    return whole_word_ids


def get_next_tokens(token_ids, index, model_type, tok):
    res = models_subwords_map.get(model_type)
    if res is None:
        return []

    # i, whole_word_ids = index+1, [token_ids[index]]
    i, whole_word_ids = index + 1, []
    spl_char, start_token_contain_spl_char = res

    while i < len(token_ids):
        curr_token = tok.convert_ids_to_tokens(token_ids[i])
        if curr_token.startswith(spl_char):
            if start_token_contain_spl_char:
                break
        else:
            if not start_token_contain_spl_char:
                break
        whole_word_ids.append(token_ids[i])
        i += 1
    return whole_word_ids


def get_results(batch_tokens, batch_labels, tokenizer, id2label, model_type, num_items):
    results = []
    for index, (tokens, labels) in enumerate(zip(batch_tokens, batch_labels)):
        labels = [id2label[x] for x in labels]
        if index >= num_items:
            break
        prev_label, token_list, entities = labels[0], [(0, tokens[0])], []
        for token_index, (token, label) in enumerate(list(zip(tokens[1:], labels[1:]))):
            label = label.split("-")[-1]
            if label == "O":
                if prev_label != "O":
                    entities.append((token_list, prev_label))
                token_list, prev_label = list(), label

            if prev_label == label:
                token_list.append((token_index + 1, token))
            else:
                if prev_label != "O":
                    entities.append((token_list, prev_label))
                token_list = list()
                prev_label = label
                token_list.append((token_index + 1, token))
        if token_list:
            entities.append((token_list, prev_label))

        # entities is a list [([(4, 1196), (5, 345)], "Address"), ([(34, 444), (35, 3834)], "Person")]
        entity_list = []
        for item in entities:
            token_list, entity = item
            # print("Before - ", tokenizer.decode([x[1] for x in token_list]))
            prev_tokens = get_previous_tokens(
                tokens, token_list[0][0], model_type, tokenizer
            )
            end_tokens = get_next_tokens(
                tokens, token_list[-1][0], model_type, tokenizer
            )

            token_list = prev_tokens + [x[1] for x in token_list[1:]] + end_tokens
            # print("After - ", tokenizer.decode(token_list))
            entity_text = process_text(
                tokenizer.decode(token_list, skip_special_tokens=True)
            )
            if len(entity_text) > 2:
                entity_list.append((entity_text, entity))

        entity_dict = {}
        text = process_text(
            tokenizer.decode(batch_tokens[index], skip_special_tokens=True)
        )
        entity_dict["Text"] = text
        _ = [entity_dict.setdefault(x[1], []).append(x[0]) for x in entity_list if x[0]]
        results.append(entity_dict)
    return results



import concurrent.futures


def _run_async(fn, **inputs):
    """runs the inputs asynchronously"""
    tp = concurrent.futures.ThreadPoolExecutor(1)
    try:
        future = tp.submit(fn=fn, **inputs)
    except:
        future = tp.submit(fn, **inputs)
    tp.shutdown(False)
    return future


class EditFeatureJob(object):
    """
    Represents a Single Editing Job.  The `EditFeatureJob` class allows for the
    asynchronous operation of the :meth:`~arcgis.features.FeatureLayer.edit_features`
    method. This class is not intended for users to initialize directly, but is
    retuned by :meth:`~arcgis.features.FeatureLayer.edit_features` when `future=True`.


    ================  ===============================================================
    **Parameter**      **Description**
    ----------------  ---------------------------------------------------------------
    future            Future. The future request.
    ----------------  ---------------------------------------------------------------
    connection        The GIS connection object.
    ================  ===============================================================

    """

    _future = None
    _con = None

    # ----------------------------------------------------------------------
    def __init__(self, future, connection):
        """
        initializer
        """
        assert isinstance(future, concurrent.futures.Future)
        self._con = connection
        self._future = future

    # ----------------------------------------------------------------------
    def __str__(self):
        return "<Edit Feature Job>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()

    # ----------------------------------------------------------------------
    @property
    def task(self):
        """Returns the task name.
        :return: string
        """
        return "Edit Features Job"

    # ----------------------------------------------------------------------
    @property
    def messages(self):
        """
        returns the GP messages

        :return: List
        """
        return self._future.result()

    # ----------------------------------------------------------------------
    @property
    def status(self):
        """
        returns the Job status

        :return: bool - True means running, False means finished
        """
        return self.running()

    # ----------------------------------------------------------------------
    def cancelled(self):
        """
        Return True if the call was successfully cancelled.

        :return: boolean
        """
        return self._future.cancelled()

    # ----------------------------------------------------------------------
    def running(self):
        """
        Return True if the call is currently being executed and cannot be cancelled.

        :return: boolean
        """
        return self._future.running()

    # ----------------------------------------------------------------------
    def done(self):
        """
        Return True if the call was successfully cancelled or finished running.

        :return: boolean
        """
        return self._future.done()

    # ----------------------------------------------------------------------
    def result(self):
        """
        Return the value returned by the call. If the call hasn't yet completed
        then this method will wait.

        :return: object
        """
        try:
            res = self._future.result()
            url = res.get("resultUrl", None)
            if url is None:
                return None
            return self._con.get(url)
        except Exception as e:
            raise e



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_base_ = "./_base_/models/faster_rcnn_r50_fpn.py"
rpn_weight = 0.7
model = dict(
    backbone=dict(
        norm_cfg=dict(requires_grad=False),
        norm_eval=True,
        style="caffe",
        init_cfg=dict(
            type="Pretrained", checkpoint="open-mmlab://detectron2/resnet50_caffe"
        ),
    ),
    rpn_head=dict(
        _delete_=True,
        type="CascadeRPNHead",
        num_stages=2,
        stages=[
            dict(
                type="StageCascadeRPNHead",
                in_channels=256,
                feat_channels=256,
                anchor_generator=dict(
                    type="AnchorGenerator",
                    scales=[8],
                    ratios=[1.0],
                    strides=[4, 8, 16, 32, 64],
                ),
                adapt_cfg=dict(type="dilation", dilation=3),
                bridged_feature=True,
                sampling=False,
                with_cls=False,
                reg_decoded_bbox=True,
                bbox_coder=dict(
                    type="DeltaXYWHBBoxCoder",
                    target_means=(0.0, 0.0, 0.0, 0.0),
                    target_stds=(0.1, 0.1, 0.5, 0.5),
                ),
                loss_bbox=dict(
                    type="IoULoss", linear=True, loss_weight=10.0 * rpn_weight
                ),
            ),
            dict(
                type="StageCascadeRPNHead",
                in_channels=256,
                feat_channels=256,
                adapt_cfg=dict(type="offset"),
                bridged_feature=False,
                sampling=True,
                with_cls=True,
                reg_decoded_bbox=True,
                bbox_coder=dict(
                    type="DeltaXYWHBBoxCoder",
                    target_means=(0.0, 0.0, 0.0, 0.0),
                    target_stds=(0.05, 0.05, 0.1, 0.1),
                ),
                loss_cls=dict(
                    type="CrossEntropyLoss",
                    use_sigmoid=True,
                    loss_weight=1.0 * rpn_weight,
                ),
                loss_bbox=dict(
                    type="IoULoss", linear=True, loss_weight=10.0 * rpn_weight
                ),
            ),
        ],
    ),
    roi_head=dict(
        bbox_head=dict(
            bbox_coder=dict(target_stds=[0.04, 0.04, 0.08, 0.08]),
            loss_cls=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.5),
            loss_bbox=dict(type="SmoothL1Loss", beta=1.0, loss_weight=1.0),
        )
    ),
    # model training and testing settings
    train_cfg=dict(
        rpn=[
            dict(
                assigner=dict(
                    type="RegionAssigner", center_ratio=0.2, ignore_ratio=0.5
                ),
                allowed_border=-1,
                pos_weight=-1,
                debug=False,
            ),
            dict(
                assigner=dict(
                    type="MaxIoUAssigner",
                    pos_iou_thr=0.7,
                    neg_iou_thr=0.7,
                    min_pos_iou=0.3,
                    ignore_iof_thr=-1,
                ),
                sampler=dict(
                    type="RandomSampler",
                    num=256,
                    pos_fraction=0.5,
                    neg_pos_ub=-1,
                    add_gt_as_proposals=False,
                ),
                allowed_border=-1,
                pos_weight=-1,
                debug=False,
            ),
        ],
        rpn_proposal=dict(max_per_img=300, nms=dict(iou_threshold=0.8)),
        rcnn=dict(
            assigner=dict(pos_iou_thr=0.65, neg_iou_thr=0.65, min_pos_iou=0.65),
            sampler=dict(type="RandomSampler", num=256),
        ),
    ),
    test_cfg=dict(
        rpn=dict(max_per_img=300, nms=dict(iou_threshold=0.8)),
        rcnn=dict(score_thr=1e-3),
    ),
)



import logging as _logging
from typing import Optional
import arcgis
from functools import lru_cache
from arcgis.gis import GIS
from arcgis.geoprocessing import import_toolbox as _import_toolbox
from arcgis._impl.common._utils import _validate_url

_log = _logging.getLogger(__name__)

_use_async = False


@lru_cache(maxsize=50)
def _create_toolbox(url, gis, verbose=False):
    """holds the cached toolbox"""
    return _import_toolbox(url_or_item=url, gis=gis, verbose=verbose)


def get_travel_modes(gis: Optional[GIS] = None):
    """


    Get a list of travel modes that can be used with directions and routing services available in your portal.

    =================================================     ========================================================================
    **Parameter**                                          **Description**
    -------------------------------------------------     ------------------------------------------------------------------------
    gis                                                   Optional, the :class:`~arcgis.gis.GIS` on which this tool runs. If not specified, the active GIS is used.
    =================================================     ========================================================================


    :return: the following as a named tuple:
    * supported_travel_modes - Supported Travel Modes as a FeatureSet
    * default_travel_mode - Default Travel Mode as a str


    See `GetTravelModes <https://developers.arcgis.com/rest/network/api-reference/gettravelmodes-synchronous-task.htm>`_ for additional help.
    """

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.routingUtilities.url
    url = _validate_url(url, gis)
    tbx = _create_toolbox(url, gis)
    return tbx.get_travel_modes()


get_travel_modes.__annotations__ = {"return": tuple}


def get_tool_info(
    service_name: str = "asyncRoute",
    tool_name: str = "FindRoutes",
    gis: Optional[GIS] = None,
):
    """


    Get additional information such as the description of the network dataset used for the analysis and the execution limits for a tool in a geoprocessing service.

    Parameters:

       service_name: Service Name (str). Required parameter.  Specify the service name containing the tool. The parameter value should be specified using one of the following keywords that reference a particular geoprocessing service.asyncClosestFacility - The asynchronous geoprocessing service used to perform the closest facility analysis.asyncLocationAllocation - The asynchronous geoprocessing service used to perform the location-allocation analysis.asyncRoute - The asynchronous geoprocessing service used to perform the route analysis.asyncServiceArea - The asynchronous geoprocessing service used to perform the service area analysis.asyncVRP - The asynchronous geoprocessing service used to perform the vehicle routing problem analysis.syncVRP - The synchronous geoprocessing service used to perform the vehicle routing problem analysis.The default value is asyncRoute.
          Choice list:['asyncClosestFacility', 'asyncLocationAllocation', 'asyncODCostMatrix', 'asyncRoute', 'asyncServiceArea', 'asyncVRP', 'syncVRP']

       tool_name: Tool Name (str). Required parameter.  Specify the tool name in the geoprocessing service. The parameter value should be a valid tool name in the geoprocessing service specified by the serviceName parameter. The default value is FindRoutes.
          Choice list:['EditVehicleRoutingProblem', 'FindClosestFacilities', 'FindRoutes', 'GenerateOriginDestinationCostMatrix', 'GenerateServiceAreas', 'SolveLocationAllocation', 'SolveVehicleRoutingProblem']

    gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


    Returns:
       tool_info - Tool Info as a str

    See https://logistics.arcgis.com/arcgis/rest/directories/arcgisoutput/World/Utilities_GPServer/World_Utilities/GetToolInfo.htm for additional help.
    """

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.routingUtilities.url
    url = _validate_url(url, gis)
    tbx = _create_toolbox(url, gis)
    kwargs = {
        "service_name": service_name,
        "tool_name": tool_name,
        "gis": gis,
        "future": False,
    }

    return tbx.get_tool_info(**kwargs)


get_tool_info.__annotations__ = {"service_name": str, "tool_name": str, "return": str}



from .kadmin import KubernetesAdmin
from ._adaptors import WebAdaptorManager
from ._architecture import ArchitectureManager
from ._content import ExternalContentManager, LanguageManager
from ._datastores import DataStores, DataStore
from ._deployment import Deployment, DeploymentManager, DeploymentProperty
from ._jobs import Job, JobManager
from ._recovery import RecoveryManager, Backup, BackupStore, BackupStoresManager
from ._license import LicenseManager
from ._logs import LogManager
from ._mode import Mode
from ._overview import Overview
from ._organizations import KubeEnterpriseGroups, KubeEnterpriseUser
from ._organizations import (
    KubeOrganization,
    KubeOrganizations,
    KubeOrgFederations,
    KubeOrgLicense,
    KubeOrgSecurity,
)
from ._security import (
    KubeSecurity,
    KubeSecurityCert,
    KubeSecurityConfig,
    KubeSecurityIngress,
    KubeSecuritySAML,
)
from ._services import KubeService, GPJobManager, ServicesManager
from ._system import (
    Container,
    Indexer,
    Server,
    ServerDefaults,
    ServerManager,
    SystemManager,
)
from ._tasks import TaskManager
from ._upgrades import UpgradeManager
from ._uploads import Uploads
from ._usage import UsageStatistics



import traceback
from .._data import _raise_fastai_import_error
from ._inference_only_models import InferenceOnlyModel

HAS_TRANSFORMER = True

try:
    import torch
    from transformers import pipeline, logging
    from fastprogress.fastprogress import progress_bar

    try:
        # For version 3.3.0
        from transformers.modeling_auto import MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING
    except ModuleNotFoundError as e:
        # For version 4.5.1
        from transformers.models.auto.modeling_auto import (
            MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING,
        )
    EXPECTED_MODEL_TYPES = [
        x.__name__.replace("Config", "")
        for x in MODEL_FOR_SEQ_TO_SEQ_CAUSAL_LM_MAPPING.keys()
    ]
except Exception as e:
    transformer_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_TRANSFORMER = False
    EXPECTED_MODEL_TYPES = []


class TextSummarizer(InferenceOnlyModel):
    """
    Creates a :class:`~arcgis.learn.text.TextSummarizer` Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    backbone                Optional string. Specify the HuggingFace
                            transformer model name which will be used to
                            summarize the text.

                            To learn more about the available models for
                            summarization task, kindly visit:-
                            https://huggingface.co/models?pipeline_tag=summarization
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    pretrained_path         Option str. Path to a directory, where pretrained
                            model files are saved.
                            If pretrained_path is provided, the model is
                            loaded from that path on the local disk.
    ---------------------   -------------------------------------------
    working_dir             Option str. Path to a directory on local filesystem.
                            If directory is not present, it will be created.
                            This directory is used as the location to save the
                            model.
    =====================   ===========================================

    :return: :class:`~arcgis.learn.text.TextSummarizer` Object
    """

    #: supported transformer architectures
    supported_backbones = EXPECTED_MODEL_TYPES

    def __init__(self, backbone=None, **kwargs):
        if not HAS_TRANSFORMER:
            _raise_fastai_import_error(import_exception=transformer_exception)
        super().__init__(backbone=backbone, task="summarization", **kwargs)

    def summarize(self, text_or_list, show_progress=True, **kwargs):
        """
        Summarize the given text or list of text

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required string or list. A text/passage
                                or a list of texts/passages to generate the
                                summary for.
        ---------------------   -------------------------------------------
        show_progress           optional Bool. If set to True, will display a
                                progress bar depicting the items processed so far.
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        min_length              Optional integer. The minimum length of the
                                sequence to be generated.
                                Default value is set to to `min_length` parameter
                                of the model config.
        ---------------------   -------------------------------------------
        max_length              Optional integer. The maximum length of the
                                sequence to be generated.
                                Default value is set to to `max_length` parameter
                                of the model config.
        ---------------------   -------------------------------------------
        num_return_sequences    Optional integer. The number of independently
                                computed returned sequences for each element
                                in the batch.
                                Default value is set to 1.
        ---------------------   -------------------------------------------
        num_beams               Optional integer. Number of beams for beam
                                search. 1 means no beam search.
                                Default value is set to 1.
        ---------------------   -------------------------------------------
        length_penalty          Optional float. Exponential penalty to the
                                length. 1.0 means no penalty. Set to values < 1.0
                                in order to encourage the model to generate
                                shorter sequences, to a value > 1.0 in order to
                                encourage the model to produce longer sequences.
                                Default value is set to 1.0.
        ---------------------   -------------------------------------------
        early_stopping          Optional bool. Whether to stop the beam search
                                when at least ``num_beams`` sentences are
                                finished per batch or not.
                                Default value is set to False.
        =====================   ===========================================

        :return: a list or a list of list containing the summary/summaries for the input prompt(s) / sentence(s)
        """
        results = []
        num_return_sequences = kwargs.get("num_return_sequences", 1)
        min_length, max_length = kwargs.get("min_length"), kwargs.get("max_length")
        if min_length and max_length and min_length > max_length:
            error_message = (
                f"Value of `min_length` parameter({min_length}) cannot be "
                f"greater than the value of `max_length` parameter({max_length})."
            )
            raise Exception(error_message)

        if not isinstance(text_or_list, (list, tuple)):
            text_or_list = [text_or_list]

        for i in progress_bar(range(len(text_or_list)), display=show_progress):
            result = self.model(text_or_list[i], **kwargs)
            if num_return_sequences == 1:
                result = result[0]
            results.append(result)
        return results



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_base_ = "./_base_/models/cascade_rcnn_r50_fpn.py"

model = dict(
    type="CascadeRCNN",
    pretrained="open-mmlab://resnext101_64x4d",
    backbone=dict(
        type="ResNeXt",
        depth=101,
        groups=64,
        base_width=4,
        num_stages=4,
        out_indices=(0, 1, 2, 3),
        frozen_stages=1,
        norm_cfg=dict(type="BN", requires_grad=True),
        style="pytorch",
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/cascade_rcnn/cascade_rcnn_x101_64x4d_fpn_20e_coco/cascade_rcnn_x101_64x4d_fpn_20e_coco_20200509_224357-051557b1.pth"



from __future__ import annotations
import concurrent.futures
from arcgis.gis import GIS, Item
from arcgis.gis.nb import NotebookManager
from arcgis.gis.agonb import AGOLNotebookManager


def _list_instances(gis: GIS = None) -> list[dict[str:str]]:
    """
    Returns a list of avaialable Machine Instances.

    :returns: list[dict[str:str]]
    """

    if gis is None:
        from arcgis import env

        gis = env.active_gis
        assert gis
    mgrs = gis.notebook_server
    if len(mgrs) > 0:
        if gis._portal.is_arcgisonline:
            mgr = mgrs[0]
            ip = mgr.instance_preferences
            return ip.instances.get("instanceTypePreferences", [])
    return []


def list_runtimes(gis: GIS = None) -> list:
    """
    Returns a list of avaialable runtimes.

    :returns: list[dict[str:str]]
    """

    if gis is None:
        from arcgis import env

        gis = env.active_gis
        assert gis
    mgrs = gis.notebook_server
    if len(mgrs) > 0:
        if gis._portal.is_arcgisonline:
            mgr = mgrs[0]
            rt = mgr.runtimes
            return rt.list()
        else:
            mgr = mgrs[0]
            nb = mgr.notebooks
            rt = nb.runtimes
            return [r.properties for r in rt]
    return []


def execute_notebook(
    item: Item,
    *,
    timeout: int = 50,
    update_portal_item: bool = True,
    parameters: list = None,
    save_parameters: bool = False,
    server_index: int = 0,
    gis: GIS = None,
    future: bool = False,
) -> dict | concurrent.futures.Future:
    """

    The Execute Notebook operation allows administrators and users with
    the `Create and Edit Notebooks` privilege to remotely run a notebook
    that they own.  The notebook specified in the operation will be run
    with all cells in order.

    Using this operation, you can schedule the execution of a notebook,
    either once or with a regular occurrence. This allows you to
    automate repeating tasks such as data collection and cleaning,
    content updates, and portal administration. On Linux machines, use
    a cron job to schedule the executeNotebook operation; on Windows
    machines, you can use the Task Scheduler app.

    .. note::
        To run this operation in ArcGIS Enterprise, you must log in with
        an Enterprise account. You cannot execute notebooks using the
        ArcGIS Notebook Server primary site administrator account.

    .. note::
        ArcGIS Online has additional parameters, as noted in the parameter
        table below.

    You can specify parameters to be used in the notebook at execution
    time. If you've specified one or more parameters, they'll be
    inserted into the notebook as a new cell. This cell will be placed
    at the beginning of the notebook, unless you have added the tag
    parameters to a cell.

    ====================    ====================================================================
    **Parameter**            **Description**
    --------------------    --------------------------------------------------------------------
    item                    Required :class:`~arcgis.gis.Item`. Opens an existing portal item.
    --------------------    --------------------------------------------------------------------
    update_portal_item      Optional Boolean. Specifies whether you want to update the
                            notebook's portal item after execution. The default is true. You may
                            want to specify true when the notebook you're executing contains
                            information that needs to be updated, such as a workflow that
                            collects the most recent version of a dataset. It may not be
                            important to update the portal item if the notebook won't store any
                            new information after executing, such as an administrative notebook
                            that emails reminders to inactive users.
    --------------------    --------------------------------------------------------------------
    parameters              Optional List. An optional array of parameters to add to the
                            notebook for this execution. The parameters will be inserted as a
                            new cell directly after the cell you have tagged ``parameters``.
                            Separate parameters with a comma. Use the format "x":1 when
                            defining parameters with numbers, and "y":"text" when defining
                            parameters with text strings.
    --------------------    --------------------------------------------------------------------
    save_parameters         Optional Boolean.  Specifies whether the notebook parameters cell
                            should be saved in the notebook for future use. The default is
                            false.
    --------------------    --------------------------------------------------------------------
    timeout                 Optional Int. The number of minutes to run the instance before timeout. This is only available on ArcGIS Online.
    --------------------    --------------------------------------------------------------------
    future                  Optional boolean. If True, a Job object will be returned and the process
                            will not wait for the task to complete. The default is False, which means wait for results.
    ====================    ====================================================================

    :returns: Dict else If ``future = True``, then the result is
              a `concurrent.futures.Future <https://docs.python.org/3/library/concurrent.futures.html>`_ object.
              Call ``result()`` to get the response

    .. code-block:: python

        #Usage example

        >>> from arcgis.gis import GIS
        >>> from arcgis.notebook import execute_notebook

        >>> gis = GIS(profile="your_org_profile")

        >>> nb_item = gis.content.get("ac7b7792913b4b3c9b22da4e2c42f986")
        >>> execute_notebook(nb_item)

    """
    if gis is None:
        from arcgis import env

        gis = env.active_gis or item._gis
        assert gis

    mgrs = gis.notebook_server
    if len(mgrs) > 0:
        if gis._portal.is_arcgisonline:
            instance_type = None
            mgr = gis.notebook_server[0]
            assert isinstance(mgr, AGOLNotebookManager)
            return mgr.notebooksmanager.execute_notebook(
                item=item,
                update_portal_item=update_portal_item,
                parameters=parameters,
                save_parameters=save_parameters,
                instance_type=instance_type,
                timeout=timeout,
                future=future,
            )
        else:
            mgr = gis.notebook_server[server_index].notebooks
            assert isinstance(mgr, NotebookManager)
            return mgr.execute_notebook(
                item=item,
                update_portal_item=update_portal_item,
                parameters=parameters,
                save_parameters=save_parameters,
                future=future,
            )

    return



from typing import Union, Optional, ClassVar
from dataclasses import dataclass

from arcgis.realtime import Velocity
from arcgis.realtime.velocity.feeds._feed_template import _FeedTemplate
from arcgis.realtime.velocity.feeds.geometry import (
    _HasGeometry,
    SingleFieldGeometry,
    XYZGeometry,
)
from arcgis.realtime.velocity.feeds.time import _HasTime, TimeInstant, TimeInterval
from arcgis.realtime.velocity.input.format import (
    EsriJsonFormat,
    GeoJsonFormat,
    JsonFormat,
    DelimitedFormat,
    XMLFormat,
    _format_from_config,
)


@dataclass
class WebSocket(_FeedTemplate, _HasTime, _HasGeometry):
    """
    Receives events from a web socket. This data class can be used to define the feed configuration and use it
    to create the feed.

    ==================          ====================================================================
    **Parameter**                **Description**
    ------------------          --------------------------------------------------------------------
    label                       String. Unique label for the feed instance.
    ------------------          --------------------------------------------------------------------
    description                 String. Feed description.
    ------------------          --------------------------------------------------------------------
    url                         String. WebSocket URL over which messages are received.
    ==================          ====================================================================

    =====================       ===============================================================================
    **Optional Argument**       **Description**
    =====================       ===============================================================================
    data_format                 [:class:`~arcgis.realtime.velocity.input.EsriJsonFormat`, :class:`~arcgis.realtime.velocity.input.GeoJsonFormat`, :class:`~arcgis.realtime.velocity.input.DelimitedFormat`, :class:`~arcgis.realtime.velocity.input.JsonFormat`, :class:`~arcgis.realtime.velocity.input.XMLFormat`].
                                An instance that contains the data format
                                configuration for this feed. Configure only allowed formats.
                                If this is not set right during initialization, a format will be
                                auto-detected and set from a sample of the incoming data. This sample
                                will be fetched from the configuration provided so far in the init.
    ---------------------       -------------------------------------------------------------------------------
    track_id_field              String. Name of the field from the incoming data that should be set as
                                track ID.
    ---------------------       -------------------------------------------------------------------------------
    geometry                    [:class:`~arcgis.realtime.velocity.feeds.XYZGeometry`, :class:`~arcgis.realtime.velocity.feeds.SingleFieldGeometry`]. An instance of geometry configuration
                                that will be used to create geometry objects from the incoming data.
    ---------------------       -------------------------------------------------------------------------------
    time                        [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration that
                                will be used to create time information from the incoming data.
    =====================       ===============================================================================

    :return: A data class with Web Socket feed configuration configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import WebSocket
        from arcgis.realtime.velocity.input.format import DelimitedFormat
        from arcgis.realtime.velocity.feeds.geometry import XYZGeometry, SingleFieldGeometry
        from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant

        web_socket = WebSocket(
            label="feed_name",
            description="feed_description",
            url = "http://feed_url.com"
        )

        # create web socket feed

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        web_socket_feed = feeds.create(web_socket)
        web_socket_feed.start()
        feeds.items

    """

    # fields that the user sets during init
    # Web Socket specific properties
    url: str

    # user can define these properties even after initialization
    data_format: Optional[
        Union[DelimitedFormat, EsriJsonFormat, GeoJsonFormat, JsonFormat, XMLFormat]
    ] = None
    # FeedTemplate properties
    track_id_field: Optional[str] = None
    # HasGeometry properties
    geometry: Optional[Union[XYZGeometry, SingleFieldGeometry]] = None
    # HasTime properties
    time: Optional[Union[TimeInstant, TimeInterval]] = None

    # FeedTemplate properties
    _name: ClassVar[str] = "websocket"

    def __post_init__(self):
        if Velocity is None:
            return
        self._util = Velocity._util

        # validation of fields
        if self._util.is_valid(self.label) == False:
            raise ValueError(
                "Label should only contain alpha numeric, _ and space only"
            )

        # generate dictionary of this feed object's properties that will be used to query test-connection and
        # sample-messages Rest endpoint
        feed_properties = self._generate_feed_properties()

        test_connection = self._util.test_connection(
            input_type="feed", payload=feed_properties
        )
        # Test connection to make sure feed can fetch schema
        if test_connection is True:
            # test connection succeeded. Now try getting sample messages
            sample_payload = {
                "properties": {
                    "maxSamplesToCollect": 5,
                    "timeoutInMillis": 5000,
                }
            }
            self._dict_deep_merge(sample_payload, feed_properties)
            # Sample messages to fetch schema/fields
            sample_messages = self._util.sample_messages(
                input_type="feed", payload=sample_payload
            )

            if sample_messages["featureSchema"] is not None:
                # sample messages succeeded. Get Feature Schema from it
                self._set_fields(sample_messages["featureSchema"])

            # if Format was not specified by user, use the auto-detected format from the sample messages response as this feed object's format.
            if self.data_format is None:
                self.data_format = _format_from_config(sample_messages)

            print(
                "Feature Schema retrieved from the Feed:",
                sample_messages["featureSchema"],
            )

            # initiate actions for each of the following properties if it was set at init
            if self.track_id_field is not None:
                self.set_track_id(self.track_id_field)
            if self.geometry is not None:
                self.set_geometry_config(self.geometry)
            if self.time is not None:
                self.set_time_config(self.time)

    def _build(self) -> dict:
        feed_configuration = {
            "id": "",
            "label": self.label,
            "description": self.description,
            "feed": {**self._generate_schema_transformation()},
            "properties": {"executable": True},
        }

        feed_properties = self._generate_feed_properties()
        self._dict_deep_merge(feed_configuration["feed"], feed_properties)
        print(feed_configuration)
        return feed_configuration

    def _generate_feed_properties(self) -> dict:
        feed_properties = {
            "name": self._name,
            "properties": {f"{self._name}.url": self.url},
        }

        if self.data_format is not None:
            format_dict = self.data_format._build()
            self._dict_deep_merge(feed_properties, format_dict)

        return feed_properties



try:
    import os, json
    import numpy as np
    import torch
    from fastai.core import split_kwargs_by_func
    import math
    from . import util
    from .._unet_utils import is_contiguous as is_cont
    from .util import (
        variable_tile_size_check,
        normalize_batch,
        fold_tensor,
        unfold_tensor,
        dihedral_transform,
        create_interpolation_mask,
        remap,
    )
    from pathlib import Path

    HAS_TORCH = True
except Exception as e:
    HAS_TORCH = False

import arcgis
from arcgis.learn import ModelExtension

try:
    import arcpy
except:
    pass


def normalize_batch_imagenetstats(batch):
    imagenet_stats = [[0.485, 0.456, 0.406], [0.229, 0.224, 0.225]]
    mean = 255 * np.array(imagenet_stats[0], dtype=np.float32)
    std = 255 * np.array(imagenet_stats[1], dtype=np.float32)

    return (batch - mean) / std


def calculate_rectangle_size_from_batch_size(batch_size):
    """
    calculate number of rows and cols to composite a rectangle given a batch size
    :param batch_size:
    :return: number of cols and number of rows
    """
    rectangle_height = int(math.sqrt(batch_size) + 0.5)
    rectangle_width = int(batch_size / rectangle_height)

    if rectangle_height * rectangle_width > batch_size:
        if rectangle_height >= rectangle_width:
            rectangle_height = rectangle_height - 1
        else:
            rectangle_width = rectangle_width - 1

    if (rectangle_height + 1) * rectangle_width <= batch_size:
        rectangle_height = rectangle_height + 1
    if (rectangle_width + 1) * rectangle_height <= batch_size:
        rectangle_width = rectangle_width + 1

    # swap col and row to make a horizontal rect
    if rectangle_height > rectangle_width:
        rectangle_height, rectangle_width = rectangle_width, rectangle_height

    if rectangle_height * rectangle_width != batch_size:
        return batch_size, 1

    return rectangle_height, rectangle_width


def convert_bounding_boxes_to_coord_list(bounding_boxes):
    """
    convert bounding box numpy array to python list of point arrays
    :param bounding_boxes: numpy array of shape [n, 4]
    :return: python array of point numpy arrays, each point array is in shape [4,2]
    """
    num_bounding_boxes = bounding_boxes.shape[0]
    bounding_box_coord_list = []
    for i in range(num_bounding_boxes):
        coord_array = np.empty(shape=(4, 2), dtype=float)
        coord_array[0][0] = bounding_boxes[i][0]
        coord_array[0][1] = bounding_boxes[i][1]

        coord_array[1][0] = bounding_boxes[i][0]
        coord_array[1][1] = bounding_boxes[i][3]

        coord_array[2][0] = bounding_boxes[i][2]
        coord_array[2][1] = bounding_boxes[i][3]

        coord_array[3][0] = bounding_boxes[i][2]
        coord_array[3][1] = bounding_boxes[i][1]

        bounding_box_coord_list.append(coord_array)

    return bounding_box_coord_list


def get_tile_size(model_height, model_width, padding, batch_height, batch_width):
    """
    Calculate request tile size given model and batch dimensions
    :param model_height:
    :param model_width:
    :param padding:
    :param batch_width:
    :param batch_height:
    :return: tile height and tile width
    """
    tile_height = (model_height - 2 * padding) * batch_height
    tile_width = (model_width - 2 * padding) * batch_width

    return tile_height, tile_width


def tile_to_batch(
    pixel_block, model_height, model_width, padding, fixed_tile_size=True, **kwargs
):
    inner_width = model_width - 2 * padding
    inner_height = model_height - 2 * padding

    band_count, pb_height, pb_width = pixel_block.shape
    pixel_type = pixel_block.dtype

    if fixed_tile_size is True:
        batch_height = kwargs["batch_height"]
        batch_width = kwargs["batch_width"]
    else:
        batch_height = math.ceil((pb_height - 2 * padding) / inner_height)
        batch_width = math.ceil((pb_width - 2 * padding) / inner_width)

    batch = np.zeros(
        shape=(batch_width * batch_height, band_count, model_height, model_width),
        dtype=pixel_type,
    )
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        # pixel block might not be the shape (band_count, model_height, model_width)
        sub_pixel_block = pixel_block[
            :,
            y * inner_height : y * inner_height + model_height,
            x * inner_width : x * inner_width + model_width,
        ]
        sub_pixel_block_shape = sub_pixel_block.shape
        batch[
            b, :, : sub_pixel_block_shape[1], : sub_pixel_block_shape[2]
        ] = sub_pixel_block

    return batch, batch_height, batch_width


def batch_to_tile(batch, batch_height, batch_width):
    batch_size, bands, inner_height, inner_width = batch.shape
    tile = np.zeros(
        shape=(bands, inner_height * batch_height, inner_width * batch_width),
        dtype=batch.dtype,
    )

    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        tile[
            :,
            y * inner_height : (y + 1) * inner_height,
            x * inner_width : (x + 1) * inner_width,
        ] = batch[b]

    return tile


def remove_bounding_boxes_in_padding(
    bounding_boxes,
    scores,
    classes,
    image_height,
    image_width,
    padding,
    batch_height=1,
    batch_width=1,
):
    """
    :param bounding_boxes: the batch of bounding boxes, shape=[B,N,4]
    :param scores: the batch of box scores, shape=[B,N]
    :param classes: the batch of labels, shape=[B,N]
    :param image_height: model height
    :param image_width: model width
    :param padding:
    :param batch_height:
    :param batch_width:
    :return:
    """
    keep_indices = np.where(
        (bounding_boxes[:, :, 0] < image_height - padding)
        & (bounding_boxes[:, :, 1] < image_width - padding)
        & (bounding_boxes[:, :, 2] > padding)
        & (bounding_boxes[:, :, 3] > padding)
    )

    inner_width = image_width - 2 * padding
    inner_height = image_height - 2 * padding

    # convert coordinates in the batch to super tile and then filter by the keep_indices
    for b in range(batch_width * batch_height):
        y = int(b / batch_width)
        x = int(b % batch_width)

        bounding_boxes[b, :, [0, 2]] = bounding_boxes[b, :, [0, 2]] + y * inner_height
        bounding_boxes[b, :, [1, 3]] = bounding_boxes[b, :, [1, 3]] + x * inner_width

    bounding_boxes = bounding_boxes[keep_indices]
    scores = scores[keep_indices]
    classes = classes[keep_indices]

    return bounding_boxes, scores, classes


class ChildObjectDetector:
    def initialize(self, model, model_as_file):
        if not HAS_TORCH:
            raise Exception(
                "Could not find the required deep learning dependencies. Ensure you have installed the required dependent libraries. See https://developers.arcgis.com/python/guide/deep-learning/"
            )

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device("cpu")
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info["ModelFile"]
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(
                os.path.join(os.path.dirname(model), model_path)
            )

        self.json_emd_file = Path(model).parent
        self.model_extension = ModelExtension.from_model(emd_path=model)
        self._learnmodel = self.model_extension
        self.model = self.model_extension.learn.model.to(self.device)
        self.model.eval()

    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    "name": "padding",
                    "dataType": "numeric",
                    "value": self.json_info["ImageHeight"] // 4,
                    "required": False,
                    "displayName": "Padding",
                    "description": "Padding",
                },
                {
                    "name": "threshold",
                    "dataType": "numeric",
                    "value": 0.5,
                    "required": False,
                    "displayName": "Confidence Score Threshold [0.0, 1.0]",
                    "description": "Confidence score threshold value [0.0, 1.0]",
                },
                {
                    "name": "nms_overlap",
                    "dataType": "numeric",
                    "value": 0.1,
                    "required": False,
                    "displayName": "NMS Overlap",
                    "description": "Maximum allowed overlap within each chip",
                },
                {
                    "name": "batch_size",
                    "dataType": "numeric",
                    "required": False,
                    "value": 64,
                    "displayName": "Batch Size",
                    "description": "Batch Size",
                },
                {
                    "name": "exclude_pad_detections",
                    "dataType": "string",
                    "required": False,
                    "domain": ("True", "False"),
                    "value": "True",
                    "displayName": "Filter Outer Padding Detections",
                    "description": "Filter detections which are outside the specified padding",
                },
            ]
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        self.padding = int(
            scalars.get("padding", self.json_info["ImageHeight"] // 4)
        )  ## Default padding Imageheight//4.
        self.nms_overlap = float(
            scalars.get("nms_overlap", 0.1)
        )  ## Default 0.1 NMS Overlap.
        self.thres = float(scalars.get("threshold", 0.5))  ## Default 0.5 threshold.
        self.batch_size = (
            int(math.sqrt(int(scalars.get("batch_size", 64)))) ** 2
        )  ## Default 64 batch_size
        self.filter_outer_padding_detections = scalars.get(
            "exclude_pad_detections", "True"
        ).lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]  ## Default value True

        (
            self.rectangle_height,
            self.rectangle_width,
        ) = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            self.rectangle_height,
            self.rectangle_width,
        )

        return {
            "extractBands": tuple(self.json_info["ExtractBands"]),
            "padding": self.padding,
            "threshold": self.thres,
            "nms_overlap": self.nms_overlap,
            "tx": tx,
            "ty": ty,
            "fixedTileSize": 1,
        }

    def vectorize(self, **pixelBlocks):  # 8 x 3 x 224 x 224
        input_image = pixelBlocks["raster_pixels"].astype(np.float32)
        batch, batch_height, batch_width = tile_to_batch(
            input_image,
            self.json_info["ImageHeight"],
            self.json_info["ImageWidth"],
            self.padding,
            fixed_tile_size=True,
            batch_height=self.rectangle_height,
            batch_width=self.rectangle_width,
        )

        if "NormalizationStats" in self.json_info:
            img_normed = normalize_batch(batch, self.json_info)
        else:
            img_normed = normalize_batch_imagenetstats(
                batch.transpose(0, 2, 3, 1)
            ).transpose(0, 3, 1, 2)

        bounding_boxes, scores, classes = detect_object(
            self.model_extension._model_conf,
            self.model,
            img_normed,
            self.device,
            batch_size=self.batch_size,
            model_info=self.json_info,
            emd_path=self.json_emd_file,
            nms_overlap=self.nms_overlap,
            thresh=self.thres,
        )

        return convert_bounding_boxes_to_coord_list(bounding_boxes), scores, classes


def detect_object(
    model_configuration,
    model,
    images,
    device,
    batch_size,
    model_info,
    emd_path,
    **kwargs
):
    tile_height, tile_width = images.shape[2], images.shape[3]
    side = math.sqrt(batch_size)
    thres = kwargs.get("thresh", 0.5)
    nms_overlap = kwargs.get("nms_overlap", 0.1)

    if "NormalizationStats" in model_info:
        transform_kwargs, kwargs = split_kwargs_by_func(
            kwargs, model_configuration.transform_input_multispectral
        )
        batch_input = model_configuration.transform_input_multispectral(
            torch.tensor(images).to(device).float(), **transform_kwargs
        )
    else:
        transform_kwargs, kwargs = split_kwargs_by_func(
            kwargs, model_configuration.transform_input
        )
        batch_input = model_configuration.transform_input(
            torch.tensor(images).to(device).float(), **transform_kwargs
        )

    pred_batch = model(batch_input)

    preds = model_configuration.post_process(
        pred_batch, nms_overlap, thres, tile_height, device
    )

    num_boxes = 0
    for batch_idx in range(batch_size):
        num_boxes = num_boxes + preds[batch_idx][0].size(0)

    bounding_boxes = np.empty(shape=(num_boxes, 4), dtype=float)
    scores = np.empty(shape=(num_boxes), dtype=float)
    classes = np.empty(shape=(num_boxes), dtype=np.uint8)

    idx = 0
    for batch_idx in range(batch_size):
        i, j = batch_idx // side, batch_idx % side

        for bbox, label, score in zip(
            preds[batch_idx][0], preds[batch_idx][1], preds[batch_idx][2]
        ):
            bbox = (bbox + 1) / 2
            bounding_boxes[idx, 0] = (bbox.data[0] + i) * tile_height
            bounding_boxes[idx, 1] = (bbox.data[1] + j) * tile_width
            bounding_boxes[idx, 2] = (bbox.data[2] + i) * tile_height
            bounding_boxes[idx, 3] = (bbox.data[3] + j) * tile_width
            scores[idx] = score.data * 100
            classes[idx] = label.data
            idx = idx + 1

    return bounding_boxes, scores, classes


class ChildImageClassifier:
    def initialize(self, model, model_as_file):
        if not HAS_TORCH:
            raise Exception(
                "Could not find the required deep learning dependencies. Ensure you have installed the required dependent libraries. See https://developers.arcgis.com/python/guide/deep-learning/"
            )

        if arcpy.env.processorType == "GPU" and torch.cuda.is_available():
            self.device = torch.device("cuda")
            arcgis.env._processorType = "GPU"
        else:
            self.device = torch.device("cpu")
            arcgis.env._processorType = "CPU"

        if model_as_file:
            with open(model, "r") as f:
                self.json_info = json.load(f)
        else:
            self.json_info = json.load(model)

        model_path = self.json_info["ModelFile"]
        if model_as_file and not os.path.isabs(model_path):
            model_path = os.path.abspath(
                os.path.join(os.path.dirname(model), model_path)
            )

        self.json_emd_file = Path(model).parent
        self.model_extension = ModelExtension.from_model(emd_path=model)
        self.model = self.model_extension.learn.model.to(self.device)
        self.model.eval()

    def getParameterInfo(self, required_parameters):
        required_parameters.extend(
            [
                {
                    "name": "padding",
                    "dataType": "numeric",
                    "value": int(self.json_info["ImageHeight"]) // 4,
                    "required": False,
                    "displayName": "Padding",
                    "description": "Padding",
                },
                {
                    "name": "batch_size",
                    "dataType": "numeric",
                    "required": False,
                    "value": 4,
                    "displayName": "Batch Size",
                    "description": "Batch Size",
                },
                {
                    "name": "test_time_augmentation",
                    "dataType": "string",
                    "required": False,
                    "value": "False"
                    if "test_time_augmentation" not in self.json_info
                    else str(self.json_info["test_time_augmentation"]),
                    "displayName": "Perform test time augmentation while predicting",
                    "description": "If True, will merge predictions from flipped and rotated images.",
                },
            ]
        )
        if self.json_info["IsEdgeDetection"]:
            required_parameters.extend(
                [
                    {
                        "name": "thinning",
                        "dataType": "string",
                        "value": "False",
                        "required": False,
                        "displayName": "thinning",
                        "description": "If True, edges will be thined to one pixel wide",
                    },
                    {
                        "name": "merge_policy",
                        "dataType": "string",
                        "required": False,
                        "value": "mean",
                        "displayName": "Policy for merging augmented predictions",
                        "description": "Policy for merging predictions('mean', 'max' or 'min'). Applicable when test_time_augmentation is True.",
                    },
                ]
            )
            if (
                self.json_info.get("ArcGISLearnVersion", False)
                and self.json_info["ArcGISLearnVersion"] > "1.8.4"
            ):
                required_parameters.extend(
                    [
                        {
                            "name": "threshold",
                            "dataType": "numeric",
                            "value": 0.5,
                            "required": False,
                            "displayName": "Confidence Score Threshold [0.0, 1.0]",
                            "description": "Confidence score threshold value [0.0, 1.0]",
                        },
                        {
                            "name": "return_probability_raster",
                            "dataType": "string",
                            "required": False,
                            "value": "False",
                            "displayName": "Return Probability Raster",
                            "description": "If True, will return the probability surface of the result.",
                        },
                    ]
                )
        else:
            required_parameters.extend(
                [
                    {
                        "name": "predict_background",
                        "dataType": "string",
                        "required": False,
                        "value": "True",
                        "displayName": "Predict Background",
                        "description": "If False, will never predict the background/NoData Class.",
                    }
                ]
            )
        required_parameters = variable_tile_size_check(
            self.json_info, required_parameters
        )
        return required_parameters

    def getConfiguration(self, **scalars):
        self.tytx = int(scalars.get("tile_size", self.json_info["ImageHeight"]))
        self.padding = int(
            scalars.get("padding", self.tytx // 4)
        )  ## Default padding Imageheight//4.
        self.batch_size = (
            int(math.sqrt(int(scalars.get("batch_size", 4)))) ** 2
        )  ## Default 4 batch_size
        self.predict_background = scalars.get("predict_background", "true").lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]  ## Default value True
        if self.json_info["IsEdgeDetection"]:
            self.thinning = scalars.get("thinning", "false").lower() in [
                "true",
                "1",
                "t",
                "y",
                "yes",
            ]
            self.thres = float(scalars.get("threshold", 0.5))  ## Default 0.5 threshold.
            self.probability_raster = scalars.get(
                "return_probability_raster", "false"
            ).lower() in ["true", "1", "t", "y", "yes"]

            self.merge_policy = scalars.get("merge_policy", "mean").lower()
        else:
            self.thinning = None
            self.thres = None
            self.probability_raster = None
            self.merge_policy = "mean"

        (
            self.rectangle_height,
            self.rectangle_width,
        ) = calculate_rectangle_size_from_batch_size(self.batch_size)
        ty, tx = get_tile_size(
            self.tytx,
            self.tytx,
            self.padding,
            self.rectangle_height,
            self.rectangle_width,
        )

        self.use_tta = scalars.get("test_time_augmentation", "false").lower() in [
            "true",
            "1",
            "t",
            "y",
            "yes",
        ]  # Default value True

        return {
            "extractBands": tuple(self.json_info["ExtractBands"]),
            "padding": self.padding,
            "tx": tx,
            "ty": ty,
            "threshold": self.thres,
            "return_probability_raster": self.probability_raster,
            "fixedTileSize": 1,
            "test_time_augmentation": self.use_tta,
            "merge_policy": self.merge_policy,
        }

    def updatePixels(self, tlc, shape, props, **pixelBlocks):  # 8 x 224 x 224 x 3
        input_image = pixelBlocks["raster_pixels"].astype(np.float32)
        batch, batch_height, batch_width = tile_to_batch(
            input_image,
            self.tytx,
            self.tytx,
            self.padding,
            fixed_tile_size=True,
            batch_height=self.rectangle_height,
            batch_width=self.rectangle_width,
        )

        class_values = [clas["Value"] for clas in self.json_info["Classes"]]
        is_contiguous = is_cont([0] + class_values)

        if not is_contiguous:
            pixel_mapping = [0] + class_values
            idx2pixel = {i: d for i, d in enumerate(pixel_mapping)}

        if "NormalizationStats" in self.json_info:
            img_normed = normalize_batch(batch, self.json_info)
        else:
            img_normed = normalize_batch_imagenetstats(
                batch.transpose(0, 2, 3, 1)
            ).transpose(0, 3, 1, 2)

        semantic_predictions = classify_image(
            self.model_extension._model_conf,
            self.model,
            img_normed,
            self.device,
            predict_bg=self.predict_background,
            model_info=self.json_info,
            emd_path=self.json_emd_file,
            thinning=self.thinning,
            threshold=self.thres,
            prob_raster=self.probability_raster,
        )

        if not is_contiguous and not self.probability_raster:
            semantic_predictions = remap(semantic_predictions, idx2pixel)

        semantic_predictions = batch_to_tile(
            semantic_predictions.cpu().numpy(), batch_height, batch_width
        )
        return semantic_predictions

    def split_predict_interpolate(self, normalized_image_tensor):
        kernel_size = self.tytx
        stride = kernel_size - (2 * self.padding)

        # Split image into overlapping tiles
        masks, t_size, patches = unfold_tensor(
            normalized_image_tensor, kernel_size, stride
        )

        if (
            self.json_info.get("ArcGISLearnVersion", "1.9.1") < "2.0.0"
        ):  # handle old models
            if "NormalizationStats" in self.json_info:
                batch_input = (
                    self.model_extension._model_conf.transform_input_multispectral(
                        patches
                    )
                )
            else:
                batch_input = self.model_extension._model_conf.transform_input(patches)

            with torch.no_grad():
                pred_batch = self.model(batch_input)

            if self.json_info["ModelName"] in ["HEDEdgeDetector", "BDCNEdgeDetector"]:
                output = pred_batch[-1]
            else:  # including MMSegmentation
                output = pred_batch

        else:
            output = classify_image(
                self.model_extension._model_conf,
                self.model,
                patches,
                self.device,
                self.predict_background,
                self.json_info,
                self.json_emd_file,
                thinning=False,
                threshold=self.thres,
                prob_raster=True,
            )

        interpolation_mask = create_interpolation_mask(
            kernel_size, 0, self.device, "hann"
        )

        output = output * interpolation_mask
        masks = masks * interpolation_mask

        # merge predictions from overlapping chips
        int_surface = fold_tensor(output, masks, t_size, kernel_size, stride)

        return int_surface

    def tta_predict(self, normalized_image_tensor, test_time_aug=True):
        all_activations = []

        transforms = [0]
        if test_time_aug:
            if self.json_info["ImageSpaceUsed"] == "MAP_SPACE":
                transforms = list(range(8))
            else:
                transforms = [
                    0,
                    2,
                ]  # no vertical flips for pixel space (oriented imagery)

        for k in transforms:
            flipped_image_tensor = dihedral_transform(normalized_image_tensor[0], k)
            int_surface = self.split_predict_interpolate(
                flipped_image_tensor.unsqueeze(0)
            )
            corrected_activation = dihedral_transform(int_surface[0], k)

            if k in [5, 6]:
                corrected_activation = dihedral_transform(int_surface[0], k).rot90(
                    2, [1, 2]
                )

            all_activations.append(corrected_activation)

        all_activations = torch.stack(all_activations)

        return all_activations

    def updatePixelsTTA(self, tlc, shape, props, **pixelBlocks):  # 8 x 224 x 224 x 3
        class_values = [clas["Value"] for clas in self.json_info["Classes"]]
        is_contiguous = is_cont([0] + class_values)

        if not is_contiguous:
            pixel_mapping = [0] + class_values
            idx2pixel = {i: d for i, d in enumerate(pixel_mapping)}

        model_info = self.json_info

        input_image = pixelBlocks["raster_pixels"].astype(np.float32)
        input_image_tensor = torch.tensor(input_image).to(self.device).float()

        prithvimod = [
            "prithvi100m_burn_scar",
            "prithvi100m_crop_classification",
            "prithvi100m_sen1floods",
            "prithvi100m",
        ]
        prithvi = True if model_info.get("Kwargs")["model"] in prithvimod else False

        if (
            prithvi
            and model_info.get("Kwargs")["model"] != "prithvi100m_crop_classification"
            and input_image_tensor.max() > 1
        ):
            input_image_tensor = input_image_tensor / 10000

        if "NormalizationStats" in model_info:
            normalized_image_tensor = normalize_batch(
                input_image_tensor.cpu(), model_info, prithvi=prithvi
            )
            normalized_image_tensor = normalized_image_tensor.float().to(
                input_image_tensor.device
            )
        else:
            from torchvision import transforms

            normalize = transforms.Normalize(
                [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
            )
            normalized_image_tensor = normalize(input_image_tensor / 255.0).unsqueeze(0)

        all_activations = self.tta_predict(
            normalized_image_tensor,
            test_time_aug=self.use_tta,
        )

        if self.merge_policy == "max":
            activations = all_activations.max(dim=0)[0]
        elif self.merge_policy == "min":
            activations = all_activations.min(dim=0)[0]
        else:
            activations = all_activations.mean(dim=0)

        if self.probability_raster:
            predictions = activations.unsqueeze(0)
        else:
            if (
                self.json_info.get("ArcGISLearnVersion", "1.9.1") < "2.0.0"
                and not self.thinning
                and (
                    self.json_info["ModelName"]
                    in ["HEDEdgeDetector", "BDCNEdgeDetector"]
                )
            ):  # handle old edge detection models that use preds[-1]
                predictions = (activations.unsqueeze(0) > self.thres).byte()
            else:
                predictions = self.model_extension._model_conf.post_process(
                    activations.unsqueeze(0), thres=self.thres, thinning=self.thinning
                )

            if not is_contiguous:
                predictions = remap(predictions, idx2pixel)

        pad = self.padding

        return predictions[0].cpu().numpy()[:, pad : -pad or None, pad : -pad or None]


def classify_image(
    model_configuration,
    model,
    images,
    device,
    predict_bg,
    model_info,
    emd_path,
    thinning,
    threshold,
    prob_raster=False,
):
    if not isinstance(images, torch.Tensor):
        images = torch.tensor(images).to(device).float()

    if "NormalizationStats" in model_info:
        batch_input = model_configuration.transform_input_multispectral(images)
    else:
        batch_input = model_configuration.transform_input(images)

    with torch.no_grad():
        pred_batch = model(batch_input)

    if thinning == None:
        preds = model_configuration.post_process(pred_batch)
        if isinstance(preds, (tuple, list)):
            return torch.stack(preds)
        return preds
    else:
        if model_info.get("ArcGISLearnVersion", "1.9.1") < "2.0.0":  # handle old models
            if prob_raster:
                preds = pred_batch[-1].detach()
            else:
                preds = model_configuration.post_process(
                    pred_batch, thres=threshold, thinning=thinning
                )
        else:
            preds = model_configuration.post_process(
                pred_batch, thres=threshold, thinning=thinning, prob_raster=prob_raster
            )
        if thinning:
            return torch.stack(preds)
        else:
            return preds



import json
import logging
from functools import lru_cache
from arcgis.gis import GIS
from arcgis import network
from arcgis._impl.common._utils import _validate_url

_log = logging.getLogger()


# -------------------------------------------------------------------------
def _gp_travel_mode(gis: GIS, travel_mode: str = None) -> str:
    """Calculates the travel mode via the GP Service"""
    output = network.analysis.get_travel_modes(gis=gis)
    if travel_mode is None:
        _log.warning("Travel mode not set, using default travel mode")
        travel_mode = output.default_travel_mode
    matches = [
        feature.attributes["TravelMode"]
        for feature in output.supported_travel_modes.features
        if (
            feature.attributes["TravelModeId"].lower() == travel_mode.lower()
            or feature.attributes["AltName"].lower() == travel_mode.lower()
            or feature.attributes["Name"].lower() == travel_mode.lower()
        )
    ]

    if len(matches) > 0:
        try:
            return json.loads(matches[0])
        except:
            return matches[0]
    else:
        _log.warning(
            f"Cannot find {travel_mode}, using default: {output.default_travel_mode}."
        )
        matches = [
            feature.attributes["TravelMode"]
            for feature in output.supported_travel_modes.features
            if (
                feature.attributes["TravelModeId"].lower()
                == output.default_travel_mode.lower()
                or feature.attributes["AltName"].lower()
                == output.default_travel_mode.lower()
                or feature.attributes["Name"].lower()
                == output.default_travel_mode.lower()
            )
        ]
        return matches[0]


# -------------------------------------------------------------------------
def _route_service_travel_modes(gis, travel_mode: str = None) -> str:
    """gets the default values from the routing service"""
    url = _validate_url(gis.properties.helperServices.route.url, gis)
    route_service = network.RouteLayer(url, gis=gis)
    modes = route_service.retrieve_travel_modes()
    if travel_mode is None:
        travel_mode = modes["defaultTravelMode"]
    fn = lambda tm: travel_mode.lower() in [tm["id"].lower(), tm["name"].lower()]
    res = list(filter(fn, modes["supportedTravelModes"]))
    if len(res) > 0:
        return json.dumps(res[0])
    else:
        travel_mode = modes["defaultTravelMode"]
        res = list(filter(fn, modes["supportedTravelModes"]))
        return json.dumps(res[0])


# -------------------------------------------------------------------------
@lru_cache(maxsize=10)
def find_travel_mode(gis: GIS, travel_mode: str = None) -> str:
    """Gets and Validate the Travel Mode for the Network Analyst Tools"""
    try:
        return _gp_travel_mode(gis, travel_mode)
    except:
        return _route_service_travel_modes(gis, travel_mode)


# -------------------------------------------------------------------------
@lru_cache(maxsize=10)
def default_travel_mode(gis: GIS) -> str:
    """Gets the default travel mode for the GIS"""
    try:
        output = network.analysis.get_travel_modes(gis=gis)
        return output.default_travel_mode
    except:
        url = _validate_url(gis.properties.helperServices.route.url, gis)
        route_service = network.RouteLayer(url, gis=gis)
        modes = route_service.retrieve_travel_modes()
        return modes["defaultTravelMode"]



import logging
import urllib.parse
from typing import Union, List
from functools import lru_cache

from cachetools import cached, TTLCache
from arcgis.auth.tools import LazyLoader
from arcgis._impl.common._isd import InsensitiveDict
from arcgis.gis import GIS

_scenemgr = LazyLoader("arcgis.mapping._scenelyrs._lyrs")
_featuremgr = LazyLoader("arcgis.features.managers")
_mapservermgr = LazyLoader("arcgis.mapping._types")
_imagemgr = LazyLoader("arcgis.raster._layer")

_log = logging.getLogger()


###########################################################################
class AGOLServerManager:
    """
    Represents a Single AGO Server

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required String. The url string to the ArcGIS Online Server
    ------------------     --------------------------------------------------------------------
    gis                    Required GIS. The connection to ArcGIS Online.
    ==================     ====================================================================

    """

    _gis = None
    _url = None
    _properties = None

    def __init__(self, url: str, gis: GIS):
        """initializer"""
        self._url = url
        self._gis = gis

    # ---------------------------------------------------------------------
    def __str__(self):
        return f"< AGOLServerManager @ {self._url} >"

    # ---------------------------------------------------------------------
    def __repr__(self):
        return f"< AGOLServerManager @ {self._url} >"

    @property
    @lru_cache(maxsize=100)
    def is_tile_server(self) -> bool:
        """
        Returns if the server if hosting tiles or not

        :returns: bool
        """
        return self._url.lower().find("/tiles/") > -1

    @property
    @cached(cache=TTLCache(maxsize=10, ttl=25))
    def properties(self) -> InsensitiveDict:
        """
        Returns the server's properties. This call is cached for 25 seconds.

        :return: Dict
        """
        resp = self._gis._con.get(self._url, {"f": "json"})
        return InsensitiveDict(resp)

    @lru_cache(maxsize=50)
    def get(
        self, name: str
    ) -> Union[
        _mapservermgr.VectorTileLayerManager,
        _imagemgr.ImageryLayerCacheManager,
        _scenemgr.SceneLayerManager,
        _featuremgr.FeatureLayerCollectionManager,
        _mapservermgr.MapImageLayerManager,
    ]:
        """
        Returns a single service manager.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required String. The name of the service.
        ==================     ====================================================================

        :returns: Union[:class:`~arcgis.mapping.VectorTileLayer`,
                        :class:`~arcgis.raster.ImageryLayerCacheManager`,
                        :class:`~arcgis.mapping.SceneLayerManager`,
                        :class:`~arcgis.features.managers.FeatureLayerCollectionManager`,
                        :class:`~arcgis.mapping.MapImageLayerManager`]
        """

        if self.is_tile_server == False:
            for service in self.services:
                if service.properties.adminServiceInfo.name.lower() == name.lower():
                    return service
        else:
            for service in self.services:
                if service.properties.name.lower() == name.lower():
                    return service

        return

    def status(self, name: str) -> str:
        """
        Returns the status of a given service by name.

        ==================     ====================================================================
        **Parameter**           **Description**
        ------------------     --------------------------------------------------------------------
        name                   Required String. The name of the service.
        ==================     ====================================================================

        :returns: string
        """
        found = False
        if self.is_tile_server == False:
            if "services" in self.properties:
                for service in self.properties["services"]:
                    if "adminServiceInfo" in service:
                        service = service["adminServiceInfo"]
                    if service["name"].lower() == name.lower():
                        found = True
                        return service["status"]
        else:
            for service in self.services:
                if service.properties.name.lower() == name.lower():
                    found = True
                    if hasattr(service, "status"):
                        return service.status
                    elif "status" in service.properties:
                        return service.properties.status
                    else:
                        return "UNKNOWN"
        if found == False:
            raise Exception("Service not found.")

    @property
    def services(self) -> list:
        """Returns the Administrative Endpoints

        :returns: list
        """
        services = []
        properties = self.properties
        if "services" in properties:
            for service in properties["services"]:
                if "adminServiceInfo" in service:
                    service = service["adminServiceInfo"]
                name = urllib.parse.quote(service["name"])
                if self.is_tile_server:
                    url = f"{self._url}/{name}/{service['type']}"
                else:
                    url = f"{self._url}/{name}.{service['type']}"
                serivce_type = service["type"].lower()
                if serivce_type == "mapserver":
                    services.append(
                        _mapservermgr.MapImageLayerManager(url=url, gis=self._gis)
                    )
                elif serivce_type == "featureserver":
                    services.append(
                        _featuremgr.FeatureLayerCollectionManager(
                            url=url, gis=self._gis
                        )
                    )

                elif serivce_type.find("vector") > -1:
                    services.append(
                        _mapservermgr.VectorTileLayerManager(url=url, gis=self._gis)
                    )
                elif serivce_type == "sceneserver":
                    services.append(_scenemgr.SceneLayerManager(url=url, gis=self._gis))
                elif serivce_type == "imageserver":
                    services.append(
                        _imagemgr.ImageryLayerCacheManager(url, gis=self._gis)
                    )
                else:
                    _log.warning(f"No manager found for service: {url}.")
        return services


###########################################################################
class AGOLServersManager:
    """
    This class allows users to work with hosted tile and feature services on
    ArcGIS Online.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    gis                    Required GIS. The connection to ArcGIS Online.
    ==================     ====================================================================


    """

    _gis = None

    def __init__(self, gis: GIS):
        """initializer"""
        if gis._portal.is_arcgisonline == False:
            raise ValueError("Invalid GIS")
        self._gis = gis

    @property
    def properties(self) -> InsensitiveDict:
        """
        Returns the properties of the server

        :returns: InsensitiveDict
        """
        return InsensitiveDict(self._urls(gis=self._gis))

    @lru_cache(maxsize=254)
    def _urls(self, gis: GIS) -> dict:
        """returns the parsed urls"""
        info = gis._registered_servers()
        tile_urls = set(info["urls"].get("tiles", {}).get("https", []))
        feature_urls = set(info["urls"].get("features", {}).get("https", []))
        tile_urls = set(info["urls"].get("tiles", {}).get("https", []))
        pid = gis.properties.id
        tile_urls = [
            f"https://{url}/tiles/{pid}/arcgis/rest/admin/services"
            for url in tile_urls
            if url not in feature_urls
        ]
        feature_urls = [
            f"https://{url}/{pid}/ArcGIS/admin/services" for url in feature_urls
        ]
        return {"tile": tile_urls, "feature": feature_urls}

    @property
    def tile_server(self) -> List[AGOLServerManager]:
        """returns a list of Tile Administrative Servers"""
        return [
            AGOLServerManager(url, gis=self._gis)
            for url in self._urls(self._gis)["tile"]
        ]

    @property
    def feature_server(self) -> List[AGOLServerManager]:
        """returns a list of Feature Administrative Servers"""
        return [
            AGOLServerManager(url, gis=self._gis)
            for url in self._urls(self._gis)["feature"]
        ]

    @lru_cache(maxsize=254)
    def list(self) -> List[AGOLServerManager]:
        """
        Returns a list of all server managers

        :returns: List[:class:`~arcgis.gis.agoserver.AGOLServerManager`]
        """
        return self.tile_server + self.feature_server
